# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
JAX base primitive.
"""

from abc import ABCMeta, abstractmethod
from functools import partial

import jax
from jax.extend import core
from jax.interpreters import xla, mlir, batching
from jax._src import dispatch

from cuquantum.lib import cudensitymat_jax


class BasePrimitive(metaclass=ABCMeta):
    """
    JAX primitive base class.
    """

    name = None

    @staticmethod
    @abstractmethod
    def abstract():
        """
        Abstract evaluation of the inner primitive.
        """
        return NotImplemented

    @classmethod
    def outer_abstract(cls, *args, **kwargs):
        """
        Abstract evaluation of the outer primitive.
        """
        return cls.abstract(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def lowering():
        """
        Lowering rule of the primitive.
        """
        return NotImplemented

    @staticmethod
    @abstractmethod
    def impl():
        """
        Primal evaluation of the primitive.
        """
        return NotImplemented

    @staticmethod
    @abstractmethod
    def batcher():
        """
        Batching rule of the primitive.
        """
        return NotImplemented


def register_primitive(cls):
    """
    Register a JAX primitive.
    """

    def name_of_wrapper_p():
        return cls.name + "_wrapper"

    inner_p = core.Primitive(cls.name)
    dispatch.prim_requires_devices_during_lowering.add(inner_p)
    inner_p.multiple_results = cls.inner_multiple_results
    inner_p.def_impl(partial(xla.apply_primitive, inner_p))
    inner_p.def_abstract_eval(cls.abstract)
    mlir.register_lowering(inner_p, cls.lowering, platform="cuda")
    cls.inner_primitive = inner_p

    outer_p = core.Primitive(name_of_wrapper_p())
    dispatch.prim_requires_devices_during_lowering.add(outer_p)
    outer_p.multiple_results = cls.outer_multiple_results
    outer_p.def_impl(cls.impl)
    outer_p.def_abstract_eval(cls.outer_abstract)
    mlir.register_lowering(
        outer_p,
        mlir.lower_fun(cls.impl, multiple_results=cls.outer_multiple_results),
        platform="cuda"
    )
    batching.primitive_batchers[outer_p] = cls.batcher
    cls.outer_primitive = outer_p


for _name, _value in cudensitymat_jax.registrations().items():
    jax.ffi.register_ffi_target(_name, _value, platform="CUDA")



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
cuDensityMat context classes.
"""

import atexit
import logging

import jax.numpy as jnp
from numpy.typing import DTypeLike

from cuquantum.bindings import cudensitymat as cudm
from nvmath.internal import typemaps

from .operator import Operator


class CudensitymatContext:
    """
    cuDensityMat library context.

    This class holds the library handle and the workspace descriptor handle, which are used for all
    operator actions. A specific operator context and state context are created for each operator,
    stored in _operator_contexts and _state_contexts respectively, and retrieved when the specific
    operator action is invoked.
    """

    _handle = None
    _workspace_desc = None
    _operator_contexts = {}  # key: operator opaque pointer
    _state_contexts = {}  # key: (purity, tuple(state_shape), batch_size, dtype_name)

    logger = logging.getLogger("cudensitymat-jax.CudensitymatContext")

    @classmethod
    def _maybe_create_handle_and_workspace(cls):
        """
        Create the handle and workspace for the context if they are not already created.
        """
        if cls._handle is None:
            cls.logger.info("Initializing CudensitymatContext")
            cls._handle = cudm.create()
            cls.logger.info(f"Created handle at {hex(cls._handle)}")

            if cls._workspace_desc is None:
                cls._workspace_desc = cudm.create_workspace(cls._handle)
                cls.logger.info(f"Created workspace descriptor at {hex(cls._workspace_desc)}")
            else:  # handle is None but workspace is not None
                raise RuntimeError("Workspace descriptor and handle should be created at the same time")
        else:
            if cls._workspace_desc is None:  # handle is not None but workspace is None
                raise RuntimeError("Workspace descriptor and handle should be created at the same time")

    @classmethod
    def maybe_create_operator_context(cls, op: Operator) -> None:
        """
        Create the OperatorContext for the operator if it does not already exist.

        Args:
            op: The operator.
        """
        cls._maybe_create_handle_and_workspace()

        if op._ptr is None or op._ptr not in cls._operator_contexts:
            op_ctx = OperatorContext(op)
            cls._operator_contexts[op._ptr] = op_ctx
            cls.logger.info(f"Created OperatorContext for operator {hex(id(op))}")

    @classmethod
    def get_operator_context(cls, op_ptr: int) -> "OperatorContext":
        """
        Get the OperatorContext for a given operator pointer.
        """
        if op_ptr not in cls._operator_contexts:
            raise RuntimeError(
                f"No OperatorContext found for operator pointer {hex(op_ptr)}. "
                "Ensure maybe_create_operator_context() was called before get_operator_context()."
            )
        return cls._operator_contexts[op_ptr]

    @classmethod
    def maybe_create_state_context(
        cls,
        purity: cudm.StatePurity,
        state_shape: tuple[int, ...],
        batch_size: int,
        dtype: DTypeLike,
    ) -> "StateContext":
        """
        Get or create the StateContext for the given key (purity, state_shape, batch_size, dtype).
        """
        dtype_name = jnp.dtype(dtype).name
        state_key = (purity, tuple(state_shape), batch_size, dtype_name)
        # State shape is with the batch dimension, so need - 1.
        if purity == cudm.StatePurity.MIXED:
            num_modes = (len(state_shape) - 1) // 2
        else:
            num_modes = len(state_shape) - 1
        space_mode_extents = state_shape[-num_modes:]
        if state_key not in cls._state_contexts:
            state_ctx = StateContext(purity, space_mode_extents, batch_size, dtype)
            cls._state_contexts[state_key] = state_ctx
            cls.logger.info(f"Created StateContext for purity={purity} state_shape={state_shape} batch_size={batch_size} dtype={dtype}")
        return cls._state_contexts[state_key]

    @classmethod
    def get_state_context(
        cls,
        purity: cudm.StatePurity,
        state_shape: tuple[int, ...],
        batch_size: int,
        dtype: DTypeLike,
    ) -> "StateContext":
        """
        Get the StateContext for the given purity, state_shape, batch_size and dtype.
        """
        dtype_name = jnp.dtype(dtype).name
        state_key = (purity, tuple(state_shape), batch_size, dtype_name)
        if state_key not in cls._state_contexts:
            raise RuntimeError(
                f"No StateContext found for purity={purity}, state_shape={state_shape}, "
                f"batch_size={batch_size}, dtype={dtype}. "
                "Ensure maybe_create_state_context() was called before get_state_context()."
            )
        return cls._state_contexts[state_key]

    @classmethod
    def free(cls):
        """
        Free opaque handles to the library.

        Note: We don't store operator objects in contexts to avoid leaking JAX tracers.
        Operator objects and their GPU handles will be cleaned up by Python's garbage
        collector when they're no longer referenced. Here we only clean up the state
        handles and workspace that are managed by contexts.
        """
        cls.logger.info("Freeing CudensitymatContext")

        # Free all state handles from state contexts
        for state_ctx in cls._state_contexts.values():
            state_ctx.free()

        # Release gradient callback function references from operator contexts
        for op_ctx in cls._operator_contexts.values():
            op_ctx.free()

        # Free workspace and library handle
        if cls._workspace_desc is not None:
            cudm.destroy_workspace(cls._workspace_desc)
            cls._workspace_desc = None

        if cls._handle is not None:
            cudm.destroy(cls._handle)
            cls._handle = None

        # Clear tracking dictionaries
        CudensitymatContext._operator_contexts.clear()
        CudensitymatContext._state_contexts.clear()


atexit.register(CudensitymatContext.free)


class OperatorContext:
    """
    Operator context.

    Holds the operator-specific C-side handle and metadata needed for operator actions.
    """

    logger = logging.getLogger("cudensitymat-jax.OperatorContext")

    def __init__(self, op: Operator) -> None:
        """
        Initialize OperatorContext.

        Args:
            op: The operator object for operator action.
        """
        self.logger.info("Initializing OperatorContext")

        # Derived attributes from operator.
        self._space_mode_extents = op.dims
        self._data_type = typemaps.NAME_TO_DATA_TYPE[op.dtype.name]
        self._compute_type = typemaps.NAME_TO_COMPUTE_TYPE[op.dtype.name]

        # Create opaque handle to the operator.
        op._create(CudensitymatContext._handle)
        self._operator = op._ptr
        self._op = op

        # Hold all gradient callback functions (f) to prevent GC while C handles are alive.
        # The traced operator passed to _create may be GC'd after JIT tracing due to async
        # GPU dispatch, so we collect the f objects here where they outlive the traced op.
        # f is accessible via callback.callback on the existing WrappedScalar/TensorGradientCallback objects.
        self._callback_fns = []
        for callback in op._coeff_grad_callbacks:
            if callback is not None:
                self._callback_fns.append(callback.callback)
        for op_term in op.op_terms:
            for callback in op_term._coeff_grad_callbacks:
                if callback is not None:
                    self._callback_fns.append(callback.callback)
            for op_prod in op_term.op_prods:
                for base_op in op_prod:
                    if base_op._grad_callback is not None:
                        self._callback_fns.append(base_op._grad_callback.callback)

    def free(self):
        """
        Destroy the operator C handle and release gradient callback function references.
        """
        self._op._destroy()
        self._op = None
        self._callback_fns.clear()


class StateContext:
    """
    State context.

    Holds the C-side handles for the input/output states (and their adjoints) used in
    operator actions.
    """

    logger = logging.getLogger("cudensitymat-jax.StateContext")

    def __init__(self,
                 purity: cudm.StatePurity,
                 space_mode_extents: tuple,
                 batch_size: int,
                 dtype: DTypeLike,
                 ) -> None:
        self.logger.info("Initializing StateContext")

        self.state_purity = purity
        self.batch_size = batch_size

        # Store for use in create_adjoint_buffers.
        self._space_mode_extents = space_mode_extents
        self._data_type = typemaps.NAME_TO_DATA_TYPE[jnp.dtype(dtype).name]

        # Create opaque handles to the input and output states.
        self._state_in = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created input state at {hex(self._state_in)}")

        self._state_out = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created output state at {hex(self._state_out)}")

        # The state adjoints are to be set in create_adjoint_buffers when backward rule is called.
        self._state_in_adj = None
        self._state_out_adj = None

    def create_adjoint_buffers(self):
        """
        Create adjoint buffers for the input and output states.

        Frees any previously allocated adjoint handles before allocating new
        ones so that repeated backward passes do not leak C-side GPU memory.
        """
        if self._state_in_adj is not None:
            cudm.destroy_state(self._state_in_adj)
            self.logger.debug(f"Destroyed stale input state adjoint at {hex(self._state_in_adj)}")
            self._state_in_adj = None

        if self._state_out_adj is not None:
            cudm.destroy_state(self._state_out_adj)
            self.logger.debug(f"Destroyed stale output state adjoint at {hex(self._state_out_adj)}")
            self._state_out_adj = None

        self._state_in_adj = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created input state adjoint at {hex(self._state_in_adj)}")

        self._state_out_adj = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created output state adjoint at {hex(self._state_out_adj)}")

    def free(self):
        """
        Free opaque handles to the library.
        """
        self.logger.info("Freeing StateContext")

        if self._state_out_adj is not None:
            cudm.destroy_state(self._state_out_adj)
            self.logger.debug(f"Destroyed output state adjoint at {hex(self._state_out_adj)}")
            self._state_out_adj = None

        if self._state_in_adj is not None:
            cudm.destroy_state(self._state_in_adj)
            self.logger.debug(f"Destroyed input state adjoint at {hex(self._state_in_adj)}")
            self._state_in_adj = None

        if self._state_out is not None:
            cudm.destroy_state(self._state_out)
            self.logger.debug(f"Destroyed output state at {hex(self._state_out)}")
            self._state_out = None

        if self._state_in is not None:
            cudm.destroy_state(self._state_in)
            self.logger.debug(f"Destroyed input state at {hex(self._state_in)}")
            self._state_in = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Utility functions for cuQuantum Python JAX.
"""

import ctypes
from dataclasses import dataclass, field

import cupy as cp
import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm


@dataclass
class BufferMetadata:
    """
    Metadata for a buffer.
    """
    indices: list[int] = field(default_factory=list)
    types: list[int] = field(default_factory=list)
    ptrs: list[int] = field(default_factory=list)
    shape_dtypes: list[jax.ShapeDtypeStruct] = field(default_factory=list)

    def __add__(self, other: "BufferMetadata") -> "BufferMetadata":
        """
        Add two BufferMetadata objects.
        """
        return BufferMetadata(
            indices=self.indices + other.indices,
            types=self.types + other.types,
            ptrs=self.ptrs + other.ptrs,    
            shape_dtypes=self.shape_dtypes + other.shape_dtypes,
        )
    
    def __len__(self) -> int:
        """
        Return the number of buffers.
        """
        return len(self.indices)


def is_vmap_traced(buf):
    """
    Check if a buffer is traced by jax.vmap.
    """
    return isinstance(buf, jax.core.Tracer) and type(buf).__name__ == "BatchTracer"


def get_vmap_depth(obj) -> int:
    """
    Return the number of nested vmap transformations active on a traced object.
    """
    depth = 0
    trace = jax.core.find_top_trace(obj)
    while hasattr(trace, "parent_trace"):
        if type(trace).__name__ == "BatchTrace":
            depth += 1
        trace = trace.parent_trace
    return depth


def is_grad_inside_vmap(obj) -> bool:
    """
    Check if an AD transformation is active inside (inner to) a vmap transformation.
    """
    found_ad = False
    trace = jax.core.find_top_trace(obj)
    while hasattr(trace, "parent_trace"):
        if type(trace).__name__ in ["LinearizeTrace", "JVPTrace"]:
            found_ad = True
        elif type(trace).__name__ == "BatchTrace" and found_ad:
            return True
        trace = trace.parent_trace
    return False


def maybe_expand_dim(bufs, ndim):
    """
    Expand to a leading batch dimension when jax.vmap is applied or when the state is non-batched.
    """
    # TODO: This function expands dimension assuming the operand_layout in lowering has the batch
    # dimension as the most major axis. In multi-vmap situation this is not true in general.
    if not is_vmap_traced(bufs[0]) and len(bufs[0].shape) in [ndim, 2 * ndim]:
        bufs = tuple([buf.reshape((1, *buf.shape)) for buf in bufs])
    return bufs


def maybe_squeeze_dim(bufs, ndim):
    """
    Squeeze the leading batch dimension when jax.vmap is applied or when the state is non-batched.
    """
    # TODO: This function squeezes dimension assuming the operand_layout in lowering has the batch
    # dimension as the most major axis. In multi-vmap situation this is not true in general.
    if not is_vmap_traced(bufs[0]) and (
        len(bufs[0].shape) in [ndim + 1, 2 * ndim + 1] and bufs[0].shape[0] == 1
    ):
        bufs = tuple([buf.reshape(*buf.shape[1:]) for buf in bufs])
    return bufs


def get_state_batch_size_and_purity(state_in_bufs, ndim):
    """
    Get the batch size and purity of the state.
    """
    # First check all state buffer shapes.
    shape = state_in_bufs[0].shape
    for buf in state_in_bufs[1:]:
        if buf.shape != shape:
            raise ValueError("All input state buffers must have the same shape.")
    if not is_vmap_traced(state_in_bufs[0]) and len(shape) not in [ndim, 2 * ndim, ndim + 1, 2 * ndim + 1]:
        raise ValueError("The dimensions of the input state do not match the dimensions of the operator.")

    # Extract batch size based on the four values ndim, 2 * ndim, ndim + 1, 2 * ndim + 1.
    if is_vmap_traced(state_in_bufs[0]):
        # Detect batch size by traversing the trace stack.
        trace = jax.core.find_top_trace(state_in_bufs[0])
        batch_size = 1
        while hasattr(trace, "parent_trace"):
            if type(trace).__name__ == "BatchTrace":
                batch_size *= trace.axis_data.size
            trace = trace.parent_trace
    else:
        if len(shape) % ndim == 0:
            batch_size = 1
        else:  # ndim + 1 or 2 * ndim + 1
            batch_size = shape[0]

    # Extract purity based on the four values ndim, 2 * ndim, ndim + 1, 2 * ndim + 1.
    if len(shape) // ndim == 1:
        purity = cudm.StatePurity.PURE
    else:  # 2 * ndim or 2 * ndim + 1
        purity = cudm.StatePurity.MIXED

    return batch_size, purity


def check_and_return_final_batch_size(state_in_bufs, state_batch_size, op_batch_size):
    """
    Check and return the final batch size.
    """
    if len({1, state_batch_size, op_batch_size}) > 2:  # the set should be either {1} or {1, N}
        raise ValueError("The batch size of the input state does not match the batch size of the operator.")
    return max(state_batch_size, op_batch_size)


def check_and_return_op_device(op: "Operator") -> jax.Device | None:
    """
    Check if all coefficients and base operators are on the same device and return the device.
    """
    device = None

    def _check_device(obj):
        nonlocal device  # tell Python to use the outer scope's device variable
        if not isinstance(obj, jax.core.Tracer):  # only check object device if it is not traced
            # Check if the object device is a GPU.
            if obj.device.platform != 'gpu':
                raise ValueError("cuQuantum Python JAX only supports GPU devices.")

            # If device is already set, check it against the current object's device.
            # Otherwise, set it to the current object's device.
            if device is not None:
                if obj.device != device:
                    raise ValueError("All objects must be on the same device.")
            else:
                device = obj.device

    for op_term, op_term_coeff in zip(op.op_terms, op.coeffs):
        _check_device(op_term_coeff)

        for op_prod, op_prod_coeff in zip(op_term.op_prods, op_term.coeffs):
            _check_device(op_prod_coeff)

            for base_op in op_prod:
                _check_device(base_op.data)

    return device


def check_and_return_state_device(state_in_bufs: tuple[jax.Array, ...]) -> jax.Device | None:
    """
    Check if all state buffers are on the same device and return the device.
    """
    device = None

    def _check_device(obj):
        nonlocal device  # tell Python to use the outer scope's device variable
        if not isinstance(obj, jax.core.Tracer):  # only check object device if it is not traced
            # Check if the object device is a GPU.
            if obj.device.platform != 'gpu':
                raise ValueError("cuQuantum Python JAX only supports GPU devices.")

            # If device is already set, check it against the current object's device.
            # Otherwise, set it to the current object's device.
            if device is not None:
                if obj.device != device:
                    raise ValueError("All objects must be on the same device.")
            else:
                device = obj.device

    for buf in state_in_bufs:
        _check_device(buf)

    return device


def detect_ad_traced_object(obj) -> bool:
    """
    Detect if an object has been AD-traced.
    """
    if isinstance(obj, jax.core.Tracer):
        # trace = jax.core.find_top_trace(obj)  # innermost trace
        trace = obj._trace  # XXX
        while hasattr(trace, "parent_trace"):
            if type(trace).__name__ in ["LinearizeTrace", "JVPTrace"]:
                return True
            trace = trace.parent_trace
    return False


def get_empty_scalar_callback():
    """
    Return an empty scalar callback for gradient attachment.
    """
    def f(*args, **kwargs):
        return
    return cudm.WrappedScalarCallback(f, cudm.CallbackDevice.GPU)


def get_scalar_assignment_callback(coeff):
    """
    Return a scalar assignment callback.
    """
    def f(t, args, storage):
        storage[:] = f.coeff[0]
    f.coeff = cp.zeros((1,), dtype=coeff.dtype)
    return cudm.WrappedScalarCallback(f, cudm.CallbackDevice.GPU)


def get_empty_tensor_callback():
    """
    Return an empty tensor callback for gradient attachment.
    """
    def f(*args, **kwargs):
        return
    return cudm.WrappedTensorCallback(f, cudm.CallbackDevice.GPU)


def get_scalar_gradient_attachment_callback(data):
    """
    Return a scalar gradient attachment callback.
    """
    def f(t, args, scalar_grad, params_grad):
        f.scalar_grad += scalar_grad

    # f.scalar_grad is created as a CuPy array to escape JAX tracing.
    if is_vmap_traced(data):
        f.scalar_grad = cp.zeros(data.val.shape, dtype=data.val.dtype)
    else:
        f.scalar_grad = cp.zeros(data.shape, dtype=data.dtype)

    grad_callback = cudm.WrappedScalarGradientCallback(f, cudm.CallbackDevice.GPU)
    return grad_callback


def get_tensor_gradient_attachment_callback(data):
    """
    Return a tensor gradient attachment callback.
    """
    def f(t, args, tensor_grad, params_grad):
        # Transpose tensor_grad so that the batch dimension is the first dimension.
        transpose_inds = (tensor_grad.ndim - 1, *range(tensor_grad.ndim - 1))
        f.tensor_grad += tensor_grad.transpose(transpose_inds)

    # f.tensor_grad is created as a CuPy array to escape JAX tracing.
    if is_vmap_traced(data):
        f.tensor_grad = cp.zeros(data.val.shape, dtype=data.val.dtype)
    else:
        f.tensor_grad = cp.zeros(data.shape, dtype=data.dtype)

    grad_callback = cudm.WrappedTensorGradientCallback(f, cudm.CallbackDevice.GPU)
    return grad_callback


def get_random_odd_pointer_and_object():
    """
    Return a random odd pointer and its ctypes object.

    Returns:
        tuple: (pointer, obj) where obj must be kept alive by the caller
               to ensure the pointer remains valid.
    """
    obj = ctypes.c_short()
    ptr = ctypes.addressof(obj) + 1
    assert ptr % 2 == 1, "Temporary pointer must be an odd number."
    return ptr, obj


def fuse_batched_inputs(
    input_batch_axes: tuple[int | None, ...],
    batched_inputs: tuple[jax.Array, ...],
    vmap_axes: tuple[int | None, ...],
    num_state_components: int,
) -> tuple[tuple[jax.Array, ...], tuple[int, ...], tuple[int, ...]]:
    """
    Fuse vmap and batch dimensions in input arguments to a JAX primitive.

    Args:
        input_batch_axes: The batch axes of the input tensors.
        batched_inputs: The input tensors.
        vmap_axes: The vmap axes of the input tensors.
        num_state_components: The number of state components.

    Returns:
        prepared_inputs: The prepared input tensors.
        batch_sizes: The sizes of the batch dimensions.
        vmap_sizes: The sizes of the vmap dimensions.
    """
    assert len(vmap_axes) == len(batched_inputs)
    assert len(input_batch_axes) == len(batched_inputs)
    
    # NOTE: Assuming 0 is the dimension to be accumulated into.
    assert all(batch_axis == 0 for batch_axis in input_batch_axes)

    prepared_inputs = []

    # Prepare inputs: handle batch/vmap axes
    for arr, batch_axis, vmap_axis in zip(batched_inputs, input_batch_axes, vmap_axes):
        # NOTE: Expand to a batch dimension to be accumulated into. This only works for single vmap.
        arr = jnp.expand_dims(arr, axis=batch_axis)
        if vmap_axis is not None:
            # Since batch_axis is hardcoded to 0, the original vmap_axis moves to one position later.
            # The new vmap axis (vmap_axis + 1) is then moved to the position after batch_axis.
            arr = jnp.moveaxis(arr, vmap_axis + 1, batch_axis + 1)
        prepared_inputs.append(arr)

    # Fuse dimensions for primitive call
    fused_inputs = []
    batch_sizes = [None] * num_state_components
    vmap_sizes = [None] * num_state_components
    for i, (arr, batch_axis) in enumerate(zip(prepared_inputs, input_batch_axes)):
        if batch_axis is not None:
            shape = list(arr.shape)
            batch_size = shape[batch_axis]  # 1 or N; primitive handles the difference
            vmap_size = shape[batch_axis + 1]
            if i < num_state_components:
                batch_sizes[i] = batch_size
                vmap_sizes[i] = vmap_size

            shape[batch_axis] = vmap_size * batch_size
            shape.pop(batch_axis + 1)
            arr = arr.reshape(shape)  # (..., batch_size * vmap_size, ...)
        fused_inputs.append(arr)

    return tuple(fused_inputs), tuple(batch_sizes), tuple(vmap_sizes)


def unfuse_batched_outputs(fused_outputs: tuple[jax.Array, ...],
                           output_batch_axes: tuple[int, ...],
                           batch_sizes: tuple[int, ...],
                           vmap_sizes: tuple[int, ...],
                           ) -> tuple[tuple[jax.Array, ...], tuple[int, ...]]:
    """
    Unfuse vmap and batch dimensions in output arguments from a JAX primitive.

    Args:
        fused_outputs: The fused output tensors.
        output_batch_axes: The batch axes of the output tensors.
        batch_sizes: The sizes of the batch dimensions.
        vmap_sizes: The sizes of the vmap dimensions.

    Returns:
        unfused_outputs: The unfused output tensors.
        output_batch_axes: The batch axes of the output tensors.
    """
    assert len(fused_outputs) == len(output_batch_axes)

    # Unfuse output dimensions
    outputs = []
    for i, (out, batch_axis) in enumerate(zip(fused_outputs, output_batch_axes)):
        # Unfuse and move vmap dimension to position 0
        shape = list(out.shape)
        shape[batch_axis : batch_axis + 1] = [batch_sizes[i], vmap_sizes[i]]
        out = out.reshape(shape)
        out = jnp.moveaxis(out, batch_axis, 0)  # move batch dimension to position 0

        # NOTE: Squeeze the original batch dimension. This only works for single vmap.
        out = jnp.squeeze(out, axis=0)
        
        outputs.append(out)

    return outputs



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Matrix operator class in cuDensityMat.
"""

import logging

import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm
from nvmath.internal import typemaps

from ..utils import (
    get_empty_tensor_callback,
    get_tensor_gradient_attachment_callback,
    detect_ad_traced_object,
    is_vmap_traced,
)


@jax.tree_util.register_pytree_node_class
class MatrixOperator:
    """
    PyTree class for cuDensityMat's matrix operator.
    """

    logger = logging.getLogger("cudensitymat-jax.MatrixOperator")

    def __init__(self, data: jax.Array) -> None:
        """
        Initialize a MatrixOperator object.

        Args:
            data: Data buffer of the matrix operator.
        """
        if isinstance(data, jax.Array):

            # Set data and batch size.
            if data.ndim % 2 == 0:
                # Expanding to a leading dimension 1 is necessary since we are taking 0 as the batch
                # dimension when passing to ffi_lowering.
                self.data = jnp.expand_dims(data, 0)
                self.batch_size = 1
            else:  # batched
                self.data = data
                self.batch_size = data.shape[0]

            # Set other attributes derived from data.
            self.num_modes: int = len(self.data.shape) // 2
            if self.data.shape[-2 * self.num_modes:-self.num_modes] != self.data.shape[-self.num_modes:]:
                raise ValueError("Data must have the same shape on the bra and ket modes.")
            self.mode_extents: tuple[int, ...] = self.data.shape[-self.num_modes:]
            self.dtype: jnp.dtype = self.data.dtype

        elif type(data) is object:  # data is object() during AD tracing.
            # Dummy variables for derived attributes.
            self.data: object = data
            self.batch_size: int = 1
            self.num_modes: int = 0
            self.mode_extents: tuple[int, ...] = ()
            self.dtype: jnp.dtype = jnp.dtype(float)

        # Callbacks and requires_grad are set in _create() when AD tracing is detected.
        self.requires_grad = None
        self._callback = None
        self._grad_callback = None
        self._grad_ptr: int = 0

        self._ptr: int | None = None
        self._is_elementary: bool = False

    def tree_flatten(self):
        """
        Flatten the matrix operator PyTree.
        """
        children = (self.data,)
        aux_data = (
            self.batch_size,
            self.num_modes,
            self.mode_extents,
            self.dtype,
            self.requires_grad,
            self._callback,
            self._grad_callback,
            self._grad_ptr,
            self._is_elementary,
            self._ptr,
        )
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the matrix operator PyTree.
        """
        inst = cls.__new__(cls)
        inst.data = children[0]
        (
            inst.batch_size,
            inst.num_modes,
            inst.mode_extents,
            inst.dtype,
            inst.requires_grad,
            inst._callback,
            inst._grad_callback,
            inst._grad_ptr,
            inst._is_elementary,
            inst._ptr,
        ) = aux_data
        return inst

    @property
    def in_axes(self) -> "MatrixOperator":
        """
        Return the in_axes PyTree spec for vmapping over the batch dimension (axis 0 of data).
        """
        _, aux_data = self.tree_flatten()
        if is_vmap_traced(self.data) or self.batch_size > 1:
            in_axes_data = 0
        else:
            in_axes_data = None
        return type(self).tree_unflatten(aux_data, (in_axes_data,))

    def _copy(self) -> "MatrixOperator":
        """
        Copy the matrix operator.
        """
        mat_op = type(self).__new__(type(self))

        mat_op.data = jnp.copy(self.data)

        mat_op.batch_size = self.batch_size
        mat_op.num_modes = self.num_modes
        mat_op.mode_extents = self.mode_extents
        mat_op.dtype = self.dtype
        mat_op.requires_grad = self.requires_grad
        mat_op._callback = self._callback
        mat_op._grad_callback = self._grad_callback
        mat_op._grad_ptr = self._grad_ptr
        mat_op._is_elementary = self._is_elementary
        mat_op._ptr = self._ptr

        return mat_op

    def _create(self, handle):
        """
        Create opaque handle to the matrix operator.
        """
        # Create opaque handle to the matrix operator.
        if self._ptr is None:

            # Detect if data requires gradient and assign callback, gradient callback.
            self.requires_grad = detect_ad_traced_object(self.data)
            if self.requires_grad:
                self._callback = get_empty_tensor_callback()
                self._grad_callback = get_tensor_gradient_attachment_callback(self.data)
                self._grad_ptr = self._grad_callback.callback.tensor_grad.data.ptr

            if self.batch_size == 1:
                self._ptr = cudm.create_matrix_operator_dense_local(
                    handle,
                    self.num_modes,
                    self.mode_extents,
                    typemaps.NAME_TO_DATA_TYPE[self.dtype.name],
                    0,  # buffer pointer to be attached in the XLA layer
                    self._callback,
                    self._grad_callback,
                )
            else:
                self._ptr = cudm.create_matrix_operator_dense_local_batch(
                    handle,
                    self.num_modes,
                    self.mode_extents,
                    self.batch_size,
                    typemaps.NAME_TO_DATA_TYPE[self.dtype.name],
                    0,  # buffer pointer to be attached in the XLA layer
                    self._callback,
                    self._grad_callback,
                )

            self.logger.debug(f"Created matrix operator at {hex(self._ptr)}")

    def _destroy(self):
        """
        Destroy opaque handle to the matrix operator.
        """
        if self._ptr is not None:
            cudm.destroy_matrix_operator(self._ptr)
            self.logger.debug(f"Destroyed matrix operator at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Elementary operator class in cuDensityMat.
"""

import logging
from collections.abc import Sequence

import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm
from nvmath.internal import typemaps

from ..utils import (
    get_empty_tensor_callback,
    get_tensor_gradient_attachment_callback,
    detect_ad_traced_object,
    is_vmap_traced,
)


@jax.tree_util.register_pytree_node_class
class ElementaryOperator:
    """
    PyTree class for cuDensityMat's elementary operator.
    """

    logger = logging.getLogger("cudensitymat-jax.ElementaryOperator")

    def __init__(self, data: jax.Array, diag_offsets: Sequence[int] = ()) -> None:
        """
        Initialize an ElementaryOperator object.

        Args:
            data: Data buffer of the elementary operator.
            diag_offsets: Diagonal offsets of the elementary operator.
        """
        # Check consistency of tensor data.
        if isinstance(data, jax.Array):

            if len(diag_offsets) == 0:  # dense elementary operator

                # Set batch size and data.
                if data.ndim % 2 == 0:
                    # Expanding to a leading dimension 1 is necessary since we are taking 0 as the batch
                    # dimension when passing to ffi_lowering.
                    self.data = jnp.expand_dims(data, 0)
                    self.batch_size = 1
                else:  # batched elementary operator
                    # TODO: Batch dimension is assumed to be dimension 0. This constraint could be
                    # relaxed in the future.
                    self.data = data
                    self.batch_size = data.shape[0]

                self.sparsity = cudm.ElementaryOperatorSparsity.OPERATOR_SPARSITY_NONE

            else:  # multidiagonal elementary operator

                # Set batch size and data.
                if data.ndim == 2:
                    # Expanding to a leading dimension 1 is necessary since we are taking 0 as the batch
                    # dimension when passing to ffi_lowering.
                    self.data = jnp.expand_dims(data, 0)
                    self.batch_size = 1
                elif data.ndim == 3:
                    # TODO: Batch dimension is assumed to be dimension 0. This constraint could be
                    # relaxed in the future.
                    self.data = data
                    self.batch_size = data.shape[0]
                else:
                    raise ValueError("Only single-mode multidiagonal elementary operator is supported.")

                # Check diagonal offsets.
                if len(diag_offsets) != len(set(diag_offsets)):
                    raise ValueError("Diagonal offsets cannot contain duplicate elements.")
                if data.shape[-1] != len(diag_offsets):
                    raise ValueError("Number of columns in data does not match length of diagonal offsets.")

                self.sparsity = cudm.ElementaryOperatorSparsity.OPERATOR_SPARSITY_MULTIDIAGONAL

            # The following attributes are set in the same way for dense and multidiagonal elementary 
            # operators. Also check that bra and ket modes have the same shape.
            self.num_modes = data.ndim // 2
            if len(diag_offsets) == 0:  # only check on dense elementary operators
                bra_modes = self.data.shape[-2 * self.num_modes:-self.num_modes]
                ket_modes = self.data.shape[-self.num_modes:]
                if bra_modes != ket_modes:
                    raise ValueError("Dense elementary operator data must have the same shape on the bra and ket modes.")
            self.mode_extents = self.data.shape[1:self.num_modes + 1]  # skip the batch dimension.
            self.dtype: jnp.dtype = self.data.dtype

        elif type(data) is object:  # data is object() during AD tracing.
            # Dummy variables for derived attributes.
            self.batch_size: int = 1
            self.data: object = data
            self.num_modes: int = 0
            self.mode_extents: tuple[int, ...] = ()
            self.dtype: jnp.dtype = jnp.dtype(float)

        # Callbacks and diagonal offsets.
        self.diag_offsets: tuple[int, ...] = tuple(diag_offsets)
        
        self.requires_grad = None
        self._callback = None
        self._grad_callback = None
        self._grad_ptr: int = 0

        self._ptr: int | None = None
        self._is_elementary: bool = True

    def tree_flatten(self):
        """
        Flatten the elementary operator PyTree.
        """
        children = (self.data,)
        aux_data = (
            self.diag_offsets,
            self.batch_size,
            self.sparsity,
            self.num_modes,
            self.mode_extents,
            self.dtype,
            self.requires_grad,
            self._callback,
            self._grad_callback,
            self._grad_ptr,
            self._is_elementary,
            self._ptr,
        )
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the elementary operator PyTree.
        """
        inst = cls.__new__(cls)
        inst.data = children[0]
        (
            inst.diag_offsets,
            inst.batch_size,
            inst.sparsity,
            inst.num_modes,
            inst.mode_extents,
            inst.dtype,
            inst.requires_grad,
            inst._callback,
            inst._grad_callback,
            inst._grad_ptr,
            inst._is_elementary,
            inst._ptr,
        ) = aux_data
        return inst

    @property
    def in_axes(self) -> "ElementaryOperator":
        """
        Return the in_axes PyTree spec for vmapping over the batch dimension (axis 0 of data).
        """
        _, aux_data = self.tree_flatten()
        if is_vmap_traced(self.data) or self.batch_size > 1:
            in_axes_data = 0
        else:
            in_axes_data = None
        return type(self).tree_unflatten(aux_data, (in_axes_data,))

    def _copy(self) -> "ElementaryOperator":
        """
        Copy the elementary operator.
        """
        elem_op = type(self).__new__(type(self))

        elem_op.data = jnp.copy(self.data)

        elem_op.diag_offsets = self.diag_offsets
        elem_op.batch_size = self.batch_size
        elem_op.sparsity = self.sparsity
        elem_op.num_modes = self.num_modes
        elem_op.mode_extents = self.mode_extents
        elem_op.dtype = self.dtype
        elem_op.requires_grad = self.requires_grad
        elem_op._callback = self._callback
        elem_op._grad_callback = self._grad_callback
        elem_op._grad_ptr = self._grad_ptr
        elem_op._is_elementary = self._is_elementary
        elem_op._ptr = self._ptr

        return elem_op

    def _create(self, handle):
        """
        Create opaque handle to the elementary operator.
        """
        # Create opaque handle to the elementary operator
        if self._ptr is None:

            # Detect if data requires gradient and assign callback, gradient callback.
            self.requires_grad = detect_ad_traced_object(self.data)
            if self.requires_grad:
                self._callback = get_empty_tensor_callback()
                self._grad_callback = get_tensor_gradient_attachment_callback(self.data)
                self._grad_ptr = self._grad_callback.callback.tensor_grad.data.ptr

            if self.batch_size == 1:
                self._ptr = cudm.create_elementary_operator(
                    handle,
                    self.num_modes,
                    self.mode_extents,
                    self.sparsity,
                    len(self.diag_offsets),
                    self.diag_offsets,
                    typemaps.NAME_TO_DATA_TYPE[self.dtype.name],
                    0,  # buffer pointer to be attached in the XLA layer
                    self._callback,
                    self._grad_callback
                )
            else:
                self._ptr = cudm.create_elementary_operator_batch(
                    handle,
                    self.num_modes,
                    self.mode_extents,
                    self.batch_size,
                    self.sparsity,
                    len(self.diag_offsets),
                    self.diag_offsets,
                    typemaps.NAME_TO_DATA_TYPE[self.dtype.name],
                    0,  # buffer pointer to be attached in the XLA layer
                    self._callback,
                    self._grad_callback,
                )

            self.logger.debug(f"Created elementary operator at {hex(self._ptr)}")

    def _destroy(self):
        """
        Destroy opaque handle to the elementary operator.
        """
        if self._ptr is not None:
            cudm.destroy_elementary_operator(self._ptr)
            self.logger.debug(f"Destroyed elementary operator at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Operator term class in cuDensityMat.
"""

import ctypes
import logging
from collections.abc import Sequence

import cupy as cp
import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm

from .elementary_operator import ElementaryOperator
from .matrix_operator import MatrixOperator
from ..utils import (
    get_scalar_assignment_callback,
    get_empty_scalar_callback,
    get_scalar_gradient_attachment_callback,
    get_random_odd_pointer_and_object,
    detect_ad_traced_object,
    is_vmap_traced,
)


@jax.tree_util.register_pytree_node_class
class OperatorTerm:
    """
    PyTree class for cuDensityMat's operator term.
    """

    logger = logging.getLogger("cudensitymat-jax.OperatorTerm")

    def __init__(self, dims: Sequence[int]) -> None:
        """
        Initialize an OperatorTerm object.

        Args:
            dims: Hilbert space dimensions.
        """
        # Input argument.
        self.dims: tuple[int, ...] = tuple(dims)

        # Attributes for handling arguments in append.
        self.op_prods: list[tuple[ElementaryOperator | MatrixOperator, ...]] = []
        self.modes: list[tuple[int, ...]] = []
        self.conjs: list[tuple[bool, ...]] = []
        self.duals: list[tuple[bool, ...]] = []

        # When coeff is stored inside, it is always a jax.Array.
        self.coeffs: list[jax.Array] = []

        # Attributes inferred from multiple append calls.
        self.batch_sizes: list[int] = []  # keep track of batch sizes of all operator products
        self.batch_size: int = 1
        self.dtype: jnp.dtype | None = None

        # Internal attributes from interfacing to cuDensityMat.
        self._op_prod_types: list[type[ElementaryOperator] | type[MatrixOperator]] = []
        self._ptr: int | None = None

        # Attributes for handling gradients.
        self._coeff_requires_grads = []
        self._coeff_callbacks = []
        self._coeff_grad_callbacks = []
        self._coeff_ptrs: list[int] = []
        self._coeff_ptr_objs: list[ctypes.c_short | None] = []  # Keep ctypes objects alive
        self._coeff_grad_ptrs: list[int] = []
        self._coeff_grad_ptr_objs: list[ctypes.c_short | None] = []  # Keep ctypes objects alive
        self._total_coeffs_ptrs: list[int] = []
        self._total_coeffs_ptr_objs: list[ctypes.c_short | None] = []  # Keep ctypes objects alive

    def tree_flatten(self):
        """
        Flatten the operator term into a PyTree.
        """
        children = (self.op_prods, self.coeffs)
        aux_data = (
            self.dims,
            self.modes,
            self.conjs,
            self.duals,
            self.batch_size,
            self.batch_sizes,
            self.dtype,
            self._op_prod_types,
            self._ptr,
            self._coeff_ptrs,
            self._coeff_ptr_objs,
            self._coeff_grad_ptrs,
            self._coeff_requires_grads,
            self._coeff_callbacks,
            self._coeff_grad_callbacks,
            self._coeff_grad_ptr_objs,
            self._total_coeffs_ptrs,
            self._total_coeffs_ptr_objs,
        )
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the operator term from a PyTree.
        """
        inst = cls.__new__(cls)
        inst.op_prods, inst.coeffs = children
        (
            inst.dims,
            inst.modes,
            inst.conjs,
            inst.duals,
            inst.batch_size,
            inst.batch_sizes,
            inst.dtype,
            inst._op_prod_types,
            inst._ptr,
            inst._coeff_ptrs,
            inst._coeff_ptr_objs,
            inst._coeff_grad_ptrs,
            inst._coeff_requires_grads,
            inst._coeff_callbacks,
            inst._coeff_grad_callbacks,
            inst._coeff_grad_ptr_objs,
            inst._total_coeffs_ptrs,
            inst._total_coeffs_ptr_objs,
        ) = aux_data
        return inst

    @property
    def in_axes(self) -> "OperatorTerm":
        """
        Return the in_axes PyTree spec for vmapping over the batch dimension.
        """
        in_axes_op_prods = [tuple(op.in_axes for op in op_prod) for op_prod in self.op_prods]
        in_axes_coeffs = []
        for coeff in self.coeffs:
            if is_vmap_traced(coeff) or len(coeff) > 1:
                in_axes_coeffs.append(0)
            else:
                in_axes_coeffs.append(None)

        _, aux_data = self.tree_flatten()
        return type(self).tree_unflatten(aux_data, (in_axes_op_prods, in_axes_coeffs))

    def _copy(self) -> "OperatorTerm":
        """
        Internal method to copy the operator term for VJP backward pass.
        """
        op_term = type(self).__new__(type(self))

        op_term.op_prods = [tuple(base_op._copy() for base_op in op_prod) for op_prod in self.op_prods]
        op_term.coeffs = [jnp.copy(c) for c in self.coeffs]

        op_term.dims = self.dims
        op_term.modes = self.modes.copy()
        op_term.conjs = self.conjs.copy()
        op_term.duals = self.duals.copy()
        op_term.batch_size = self.batch_size
        op_term.batch_sizes = self.batch_sizes.copy()
        op_term.dtype = self.dtype
        op_term._op_prod_types = self._op_prod_types.copy()
        op_term._ptr = self._ptr
        op_term._coeff_ptrs = self._coeff_ptrs.copy()
        op_term._coeff_ptr_objs = self._coeff_ptr_objs.copy()
        op_term._coeff_grad_ptrs = self._coeff_grad_ptrs.copy()
        op_term._coeff_grad_ptr_objs = self._coeff_grad_ptr_objs.copy()
        op_term._coeff_requires_grads = self._coeff_requires_grads.copy()
        op_term._coeff_callbacks = self._coeff_callbacks.copy()
        op_term._coeff_grad_callbacks = self._coeff_grad_callbacks.copy()
        op_term._total_coeffs_ptrs = self._total_coeffs_ptrs.copy()
        op_term._total_coeffs_ptr_objs = self._total_coeffs_ptr_objs.copy()

        return op_term

    def _check_and_set_dtype(self, op_prod) -> None:
        """
        Check if the operator product has the same data type as the operator term.
        """
        if self.dtype is None:
            # If the data type is not set, set it to the data type of the first operator.
            self.dtype = op_prod[0].dtype
            for op in op_prod[1:]:
                if op.dtype != self.dtype:
                    raise ValueError("All elementary or matrix operators must have the same data type.")
        else:
            # If the data type is set, check if all elementary or matrix operators
            # have the same data type.
            for op in op_prod:
                if op.dtype != self.dtype:
                    raise ValueError("All elementary or matrix operators must have the same data type.")

    def _check_and_append_op_prod_type(self, op_prod):
        """
        Check if all terms in an operator product are of the same type.
        """
        op_prod_type = type(op_prod[0])
        for op in op_prod[1:]:
            if not isinstance(op, op_prod_type):
                raise ValueError("All terms in an operator product must be of the same type.")
        self._op_prod_types.append(op_prod_type)

    def _check_and_set_batch_size(self,
                                  op_prod: Sequence[ElementaryOperator | MatrixOperator],
                                  coeff: jax.Array,
                                  ) -> None:
        """
        Check if the operator product and coefficient batch sizes are consistent.
        """
        # Extract the batch size for the operator product.
        op_prod_batch_size = max([base_op.batch_size for base_op in op_prod])
        for base_op in op_prod:
            if base_op.batch_size not in (1, op_prod_batch_size):
                raise ValueError("All basic operators in an operator product must have batch size 1 or N.")

        # Possibly update the batch size of this operator term and check consistency.
        batch_size = max(op_prod_batch_size, len(coeff))
        if self.batch_size == 1:
            self.batch_size = batch_size
        else:
            if batch_size not in (1, self.batch_size):
                raise ValueError("Batch size in this operator product does not match batch size of this operator term.")

    def append(self,
               op_prod: Sequence[ElementaryOperator | MatrixOperator],
               modes: Sequence[int] | None = None,
               conjs: Sequence[bool] | None = None,
               duals: Sequence[bool] | None = None,
               coeff: float | complex | jax.Array = 1.0,
               ) -> None:
        """
        Append an elementary or matrix product to an operator term.

        Args:
            op_prod: Product of elementary or matrix operators to be appended.
            modes: Modes acted on by the operator product.
            conjs: Conjugations in the operator product. Only applies to MatrixOperators.
            duals: Dualities of the operator product.
            coeff: Non-batched coefficient or batched coefficients of the operator product.
        """
        if self._ptr is not None:
            raise RuntimeError("Cannot modify operator term after it has been used in an operator action.")

        # coeff is converted to a length-1 array if it is a Python scalar.
        if (
            isinstance(coeff, (float, complex)) or
            (isinstance(coeff, jax.Array) and coeff.ndim == 0)  # scalar but traced
        ):
            coeff = jnp.array([coeff], dtype=jnp.complex128)
        elif isinstance(coeff, jax.Array) and coeff.ndim > 0:
            if coeff.dtype != jnp.complex128:
                raise ValueError("Coefficient must be of type complex128.")
        else:
            raise ValueError("Coefficient must be a float, complex, or jax.Array.")

        # Check if all elementary or matrix operators have the same data type.
        self._check_and_set_dtype(op_prod)
        self._check_and_append_op_prod_type(op_prod)
        self._check_and_set_batch_size(op_prod, coeff)

        # Check consistency and append modes, conjs and duals.
        if self._op_prod_types[-1] is ElementaryOperator:
            # Modes have to specified for elementary operators.
            if modes is None:
                raise ValueError("Modes acted on must be specified for elementary operators.")

            # Check all modes are in Hilbert space.
            if not set(modes) <= set(range(len(self.dims))):
                raise ValueError("Modes acted on must be in the Hilbert space, i.e. between 0 and len(self.dims) - 1")

            # Check length of modes acted on are the same as combined number of modes in the operator product.
            if len(modes) != sum([elem_op.num_modes for elem_op in op_prod]):
                raise ValueError(f"Number of modes acted on {len(modes)} does not match combined number of modes in the operator product.")

            # Check mode extents of each elementary operator match corresponding qubit dimensions.
            modes_index = 0
            for elem_op in op_prod:
                if elem_op.mode_extents != tuple(
                    [self.dims[modes[i]] for i in range(modes_index, modes_index + elem_op.num_modes)]
                ):
                    raise ValueError("Mode extents of each elementary operator must match corresponding qubit dimensions.")
                modes_index += elem_op.num_modes

            # Check that matrix conjugations cannot be specified for elementary operators.
            if conjs is not None:
                raise ValueError("Matrix conjugations cannot be specified for elementary operators.")

            # Check that number of duals matches number of modes.
            if duals is None:
                duals = (False,) * len(modes)
            else:
                if len(duals) != len(modes):
                    raise ValueError("Number of duals must match number of modes acted on for elementary operator product.")

            # For elementary operator product, we only need modes and duals.
            self.modes.append(tuple(modes))
            self.conjs.append(())  # empty tuple is appended here to preserve length
            self.duals.append(tuple(duals))

        else:  # matrix operator product
            # Check that mode extents match Hilbert space dimensions.
            for matrix_op in op_prod:
                if matrix_op.mode_extents != self.dims:
                    raise ValueError("Mode extents must match Hilbert space dimensions for matrix operators.")

            # Check that modes acted on cannot be specified for matrix operators.
            if modes is not None:
                raise ValueError("Modes acted on cannot be specified for matrix operators.")

            # Check consistency of conjs.
            if conjs is None:
                conjs = (False,) * len(op_prod)
            else:
                if len(conjs) != len(op_prod):
                    raise ValueError("Number of matrix conjugations must match number of operator products.")

            # Check that number of duals matches number of matrix operators.
            if duals is None:
                duals = (False,) * len(op_prod)
            else:
                if len(duals) != len(op_prod):
                    raise ValueError("Number of duals must match number of matrix operators.")

            # For matrix operator product, we only need conjs and duals.
            self.modes.append(tuple(range(len(self.dims))))  # used in reference implementation during testing
            self.conjs.append(tuple(conjs))
            self.duals.append(tuple(duals))

        # Attributes from function arguments.
        self.op_prods.append(tuple(op_prod))
        self.coeffs.append(coeff)

        # Set batch size.
        self.batch_sizes.append(len(coeff))

        # Attributes for handling gradients. None is appended here to preserve length, which is then
        # updated in the _create method.
        self._coeff_requires_grads.append(None)
        self._coeff_ptrs.append(0)
        self._coeff_ptr_objs.append(None)
        self._coeff_grad_ptrs.append(0)
        self._coeff_callbacks.append(None)
        self._coeff_grad_callbacks.append(None)
        self._coeff_grad_ptr_objs.append(None)
        self._total_coeffs_ptrs.append(0)
        self._total_coeffs_ptr_objs.append(None)

    def __getitem__(self, index: int) -> tuple[ElementaryOperator | MatrixOperator, ...]:
        """
        Get an operator product from the operator term.
        """
        return self.op_prods[index]

    def _create(self, handle):
        """
        Create opaque handle to the operator term.
        """
        # Create opaque handle to dependent elementary or matrix operators.
        # XXX: We are creating redundant elementary operators here.
        # When JAX flattens and unflattens the PyTrees, it creates new Python objects.
        for op_prod in self.op_prods:
            for elem_op in op_prod:
                elem_op._create(handle)

        # Create the current operator term.
        if self._ptr is None:
            self._ptr = cudm.create_operator_term(handle, len(self.dims), self.dims)
            self.logger.debug(f"Created operator term at {hex(self._ptr)}")

            for i in range(len(self.op_prods)):

                # Detect if the coefficient requires gradient and assign callback, gradient callback,
                # temporary coefficient pointer and object.
                self._coeff_requires_grads[i] = detect_ad_traced_object(self.coeffs[i])
                if not is_vmap_traced(self.coeffs[i]) and len(self.coeffs[i]) == 1:  # non-batched coefficient

                    # Traced scalars need to be passed through an intermediate memory slot.
                    self._coeff_callbacks[i] = get_scalar_assignment_callback(self.coeffs[i])
                    self._coeff_ptrs[i] = self._coeff_callbacks[i].callback.coeff.data.ptr

                    # If gradient is computed on the coefficient, assign gradient callback and pointer.
                    if self._coeff_requires_grads[i]:
                        self._coeff_grad_callbacks[i] = get_scalar_gradient_attachment_callback(self.coeffs[i])
                        self._coeff_grad_ptrs[i] = self._coeff_grad_callbacks[i].callback.scalar_grad.data.ptr

                else:  # batched coefficients
                    if is_vmap_traced(self.coeffs[i]):
                        coeff_shape = self.coeffs[i].val.shape
                        coeff_dtype = self.coeffs[i].val.dtype
                    else:
                        coeff_shape = self.coeffs[i].shape
                        coeff_dtype = self.coeffs[i].dtype
                    static_coeff_buf = cp.ones(coeff_shape, dtype=coeff_dtype)
                    self._coeff_ptrs[i] = static_coeff_buf.data.ptr
                    self._coeff_ptr_objs[i] = static_coeff_buf

                    # For batched coefficients, callback is required by the API but is a no-op.
                    self._coeff_callbacks[i] = get_empty_scalar_callback()
                    self._total_coeffs_ptrs[i], self._total_coeffs_ptr_objs[i] = get_random_odd_pointer_and_object()

                    if self._coeff_requires_grads[i]:
                        self._coeff_grad_callbacks[i] = get_scalar_gradient_attachment_callback(self.coeffs[i])
                        self._coeff_grad_ptrs[i] = self._coeff_grad_callbacks[i].callback.scalar_grad.data.ptr

                if self._op_prod_types[i] is ElementaryOperator:
                    if self.batch_sizes[i] == 1:
                        cudm.operator_term_append_elementary_product(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [elem_op._ptr for elem_op in self.op_prods[i]],
                            self.modes[i],
                            self.duals[i],
                            1.0,  # coefficient, to be updated to the real coefficient by callback
                            self._coeff_callbacks[i],
                            self._coeff_grad_callbacks[i],
                        )
                    else:
                        cudm.operator_term_append_elementary_product_batch(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [elem_op._ptr for elem_op in self.op_prods[i]],
                            self.modes[i],
                            self.duals[i],
                            self.batch_sizes[i],
                            self._coeff_ptrs[i],
                            self._total_coeffs_ptrs[i],
                            self._coeff_callbacks[i],
                            self._coeff_grad_callbacks[i],
                        )
                else:  # MatrixOperator
                    if self.batch_sizes[i] == 1:
                        cudm.operator_term_append_matrix_product(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [mat_op._ptr for mat_op in self.op_prods[i]],
                            self.conjs[i],
                            self.duals[i],
                            1.0,  # coefficient, to be updated to the real coefficient by callback
                            self._coeff_callbacks[i],
                            self._coeff_grad_callbacks[i],
                        )
                    else:
                        cudm.operator_term_append_matrix_product_batch(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [mat_op._ptr for mat_op in self.op_prods[i]],
                            self.conjs[i],
                            self.duals[i],
                            self.batch_sizes[i],
                            self._coeff_ptrs[i],
                            self._total_coeffs_ptrs[i],
                            self._coeff_callbacks[i],
                            self._coeff_grad_callbacks[i],
                        )

            self.logger.debug(f"Appended operator products to operator term at {hex(self._ptr)}")

    def _destroy(self):
        """
        Destroy opaque handle to the operator term.
        """
        if self._ptr is not None:
            cudm.destroy_operator_term(self._ptr)
            self.logger.debug(f"Destroyed operator term at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
from functools import partial
from collections.abc import Sequence

import jax

from cuquantum.lib.cudensitymat_jax import InputType, OutputType

from .pysrc.context import CudensitymatContext
from .pysrc.operator import Operator
from .pysrc.operator_action_prim import (
    operator_action_prim,
    operator_action_backward_diff_prim,
)
from .utils import (
    BufferMetadata,
    maybe_expand_dim,
    maybe_squeeze_dim,
    get_state_batch_size_and_purity,
    check_and_return_final_batch_size,
    check_and_return_op_device,
    check_and_return_state_device,
    is_vmap_traced,
    get_vmap_depth,
    is_grad_inside_vmap,
)


logger = logging.getLogger("cudensitymat-jax.operator_action")


def operator_action(op: Operator,
                    state_in_bufs: jax.Array | Sequence[jax.Array],
                    device: jax.Device | None = None,
                    ) -> jax.Array | list[jax.Array]:
    """
    Compute the action of an operator on a state.

    Args:
        op: Operator to compute the action of.
        state_in_bufs: Buffers of the input state components.
        device: Device to use for the operator action.

    Returns:
        Buffers of the output state components.
    """
    logger.info("Calling operator_action")

    # Process input arguments.
    if isinstance(state_in_bufs, jax.Array):
        state_in_bufs = (state_in_bufs,)
    else:
        state_in_bufs = tuple(state_in_bufs)

    # Guard against nested vmap transformations, which are not supported.
    if get_vmap_depth(state_in_bufs[0]) > 1:
        raise NotImplementedError("operator_action does not support nested vmap transformations.")

    # Guard against grad applied inside vmap, which is not supported.
    if is_grad_inside_vmap(state_in_bufs[0]):
        raise NotImplementedError("operator_action does not support grad transformations inside vmap.")

    # Check and set device from op and state.
    op_device = check_and_return_op_device(op)
    state_device = check_and_return_state_device(state_in_bufs)
    if op_device is not None and state_device is not None:
        if op_device != state_device:
            raise ValueError("Operator and state buffers must be on the same device.")
        device = op_device  # set device to the common device
    else:  # if one of them is None, set device to the one that is not None
        devices = {op_device, state_device}
        devices.remove(None)
        if len(devices) > 0:  # set device only when one of them is not None
            device = devices.pop()

    # If device is still None, as in the case of tracing, set it to the first GPU device.
    if device is None:
        devices = jax.devices('gpu')
        if len(devices) == 0:
            raise ValueError("No GPU devices found.")
        device = devices[0]
        logger.info("No device specified, using the first GPU device.")

    # Check state shape and maybe expand to a leading batch dimension.
    state_batch_size, purity = get_state_batch_size_and_purity(state_in_bufs, len(op.dims))
    batch_size = check_and_return_final_batch_size(state_in_bufs, state_batch_size, op.batch_size)

    state_in_bufs = maybe_expand_dim(state_in_bufs, len(op.dims))

    # Prepare library context for forward operator action.
    # NOTE: Assuming a single state component.
    if len(state_in_bufs) > 1:
        raise NotImplementedError("More than one state component is not implemented.")

    if is_vmap_traced(state_in_bufs[0]):
        state_shape = state_in_bufs[0].val.shape
    else:
        state_shape = state_in_bufs[0].shape

    CudensitymatContext.maybe_create_operator_context(op)
    CudensitymatContext.maybe_create_state_context(purity, state_shape, batch_size, state_in_bufs[0].dtype)

    # Create metadata objects for the other inputs.
    op_term_coeff_metadata = BufferMetadata()
    op_prod_coeff_metadata = BufferMetadata()
    base_op_metadata = BufferMetadata()

    # Create metadata objects for the gradient outputs.
    op_term_coeff_grad_metadata = BufferMetadata()
    op_prod_coeff_grad_metadata = BufferMetadata()
    base_op_grad_metadata = BufferMetadata()

    # Extract temporary batched coefficient buffers for operator terms.
    for i, (
        op_term,
        op_term_coeff,
        op_term_coeff_ptr,
        op_term_coeff_grad_ptr,
        op_term_total_coeffs_ptr,
    ) in enumerate(zip(
        op.op_terms,
        op.coeffs,
        op._coeff_ptrs,
        op._coeff_grad_ptrs,
        op._total_coeffs_ptrs,
        strict=True
    )):
        is_op_term_coeff_batched = is_vmap_traced(op_term_coeff) or len(op_term_coeff) > 1
        if is_op_term_coeff_batched:
            dynamic_ptr = op_term_total_coeffs_ptr
            dynamic_type = InputType.OPERATOR_TERM_BATCHED_COEFFS.value
            if dynamic_ptr == 0:
                raise RuntimeError("Missing total coefficient pointer for batched operator term coefficient.")
        else:
            dynamic_ptr = op_term_coeff_ptr
            dynamic_type = InputType.NON_BATCHED_COEFFS.value

        if dynamic_ptr not in op_term_coeff_metadata.ptrs:
            op_term_coeff_metadata.indices.append(i)
            op_term_coeff_metadata.types.append(dynamic_type)
            op_term_coeff_metadata.ptrs.append(dynamic_ptr)

        if op_term_coeff_grad_ptr != 0:
            op_term_coeff_grad_metadata.indices.append(i)
            op_term_coeff_grad_metadata.types.append(OutputType.GRADIENT.value)
            op_term_coeff_grad_metadata.ptrs.append(op_term_coeff_grad_ptr)
            if is_vmap_traced(op_term_coeff):
                shape_dtype = jax.ShapeDtypeStruct(op_term_coeff.val.shape, op_term_coeff.val.dtype)
            else:
                shape_dtype = jax.ShapeDtypeStruct(op_term_coeff.shape, op_term_coeff.dtype)
            op_term_coeff_grad_metadata.shape_dtypes.append(shape_dtype)

        # Extract temporary batched coefficient buffers for operator products.
        for j, (
            op_prod,
            op_prod_coeff,
            op_prod_coeff_ptr,
            op_prod_coeff_grad_ptr,
            op_prod_total_coeffs_ptr,
        ) in enumerate(zip(
            op_term.op_prods,
            op_term.coeffs,
            op_term._coeff_ptrs,
            op_term._coeff_grad_ptrs,
            op_term._total_coeffs_ptrs,
            strict=True,
        )):
            is_op_prod_coeff_batched = is_vmap_traced(op_prod_coeff) or len(op_prod_coeff) > 1
            if is_op_prod_coeff_batched:
                dynamic_ptr = op_prod_total_coeffs_ptr
                dynamic_type = InputType.OPERATOR_PRODUCT_BATCHED_COEFFS.value
                if dynamic_ptr == 0:
                    raise RuntimeError("Missing total coefficient pointer for batched operator product coefficient.")
            else:
                dynamic_ptr = op_prod_coeff_ptr
                dynamic_type = InputType.NON_BATCHED_COEFFS.value

            if dynamic_ptr not in op_prod_coeff_metadata.ptrs:
                op_prod_coeff_metadata.indices.append((i, j))
                op_prod_coeff_metadata.types.append(dynamic_type)
                op_prod_coeff_metadata.ptrs.append(dynamic_ptr)

            if op_prod_coeff_grad_ptr != 0:
                op_prod_coeff_grad_metadata.indices.append((i, j))
                op_prod_coeff_grad_metadata.types.append(OutputType.GRADIENT.value)
                op_prod_coeff_grad_metadata.ptrs.append(op_prod_coeff_grad_ptr)
                if is_vmap_traced(op_prod_coeff):
                    shape_dtype = jax.ShapeDtypeStruct(op_prod_coeff.val.shape, op_prod_coeff.val.dtype)
                else:
                    shape_dtype = jax.ShapeDtypeStruct(op_prod_coeff.shape, op_prod_coeff.dtype)
                op_prod_coeff_grad_metadata.shape_dtypes.append(shape_dtype)

            # Assign certain static attributes to the operator action context.
            for k, base_op in enumerate(op_prod):
                # Checking whether base_op._ptr is None since only the unique base operators need
                # to be buffer-attached.
                if base_op._ptr is not None and base_op._ptr not in base_op_metadata.ptrs:
                    base_op_metadata.indices.append((i, j, k))
                    base_op_metadata.types.append(
                        InputType.ELEMENTARY_OPERATOR.value if base_op._is_elementary
                        else InputType.MATRIX_OPERATOR.value
                    )
                    base_op_metadata.ptrs.append(base_op._ptr)

                if base_op._grad_ptr != 0 and base_op._grad_ptr not in base_op_grad_metadata.ptrs:
                    base_op_grad_metadata.indices.append((i, j, k))
                    base_op_grad_metadata.types.append(OutputType.GRADIENT.value)
                    base_op_grad_metadata.ptrs.append(base_op._grad_ptr)
                    if is_vmap_traced(base_op.data):
                        shape_dtype = jax.ShapeDtypeStruct(base_op.data.val.shape, base_op.data.val.dtype)
                    else:
                        shape_dtype = jax.ShapeDtypeStruct(base_op.data.shape, base_op.data.dtype)
                    base_op_grad_metadata.shape_dtypes.append(shape_dtype)

    # Combine types and pointers in the same order as they are unpacked in primitive wrappers.
    other_in_metadata = op_term_coeff_metadata + op_prod_coeff_metadata + base_op_metadata
    other_out_metadata = op_term_coeff_grad_metadata + op_prod_coeff_grad_metadata + base_op_grad_metadata

    num_state_components = len(state_in_bufs)

    # Invoke operator action.
    state_out_bufs = _operator_action(
        op,
        state_in_bufs,
        device,
        batch_size,
        num_state_components,
        purity,
        tuple(other_in_metadata.types),
        tuple(other_in_metadata.ptrs),
        tuple(other_out_metadata.shape_dtypes),
        tuple(other_out_metadata.types),
        tuple(other_out_metadata.ptrs),
        tuple(op_term_coeff_metadata.indices),
        tuple(op_prod_coeff_metadata.indices),
        tuple(base_op_metadata.indices),
        tuple(op_term_coeff_grad_metadata.indices),
        tuple(op_prod_coeff_grad_metadata.indices),
        tuple(base_op_grad_metadata.indices),
    )

    # Undo the leading batch dim when it was added by maybe_expand_dim (single-state, non-vmap).
    state_out_bufs = maybe_squeeze_dim(state_out_bufs, len(op.dims))

    # Process output argument.
    if len(state_out_bufs) == 1:
        state_out_bufs = state_out_bufs[0]

    return state_out_bufs


@partial(jax.custom_vjp, nondiff_argnums=(2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16))
def _operator_action(op: Operator,
                     state_in_bufs: tuple[jax.Array, ...],
                     device: jax.Device,
                     batch_size: int,
                     num_state_components: int,
                     purity,
                     other_in_types: tuple[int, ...],
                     other_in_ptrs: tuple[int, ...],
                     other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                     other_out_types: tuple[int, ...],
                     other_out_ptrs: tuple[int, ...],
                     op_term_coeffs_indices: tuple[int, ...],
                     op_prod_coeffs_indices: tuple[int, ...],
                     base_op_indices: tuple[int, ...],
                     op_term_coeff_grad_indices: tuple[int, ...],
                     op_prod_coeff_grad_indices: tuple[tuple[int, int], ...],
                     base_op_grad_indices: tuple[tuple[int, int, int], ...],
                     ) -> list[jax.Array]:
    """
    Custom VJP rule for operator_action.
    """
    logger.info("Calling _operator_action")
    state_out_bufs, _ = _operator_action_fwd(
        op,
        state_in_bufs,
        device,
        batch_size,
        num_state_components,
        purity,
        other_in_types,
        other_in_ptrs,
        other_out_shape_dtypes,
        other_out_types,
        other_out_ptrs,
        op_term_coeffs_indices,
        op_prod_coeffs_indices,
        base_op_indices,
        op_term_coeff_grad_indices,
        op_prod_coeff_grad_indices,
        base_op_grad_indices,
    )
    return state_out_bufs


def _operator_action_fwd(op: Operator,
                         state_in_bufs: tuple[jax.Array, ...],
                         device: jax.Device,
                         batch_size: int,
                         num_state_components: int,
                         purity,
                         other_in_types: tuple[int, ...],
                         other_in_ptrs: tuple[int, ...],
                         other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                         other_out_types: tuple[int, ...],
                         other_out_ptrs: tuple[int, ...],
                         op_term_coeffs_indices: tuple[int, ...],
                         op_prod_coeffs_indices: tuple[int, ...],
                         base_op_indices: tuple[int, ...],
                         op_term_coeff_grad_indices: tuple[int, ...],
                         op_prod_coeff_grad_indices: tuple[tuple[int, int], ...],
                         base_op_grad_indices: tuple[tuple[int, int, int], ...],
                         ) -> tuple[list[jax.Array], tuple[Operator, tuple[jax.Array, ...]]]:
    """
    Forward rule for operator_action.
    """
    logger.info("Calling _operator_action_fwd")
    state_out_bufs = operator_action_prim(
        op,
        state_in_bufs,
        device,
        batch_size,
        num_state_components,
        purity,
        other_in_types,
        other_in_ptrs,
        other_out_shape_dtypes,
        other_out_types,
        other_out_ptrs,
        op_term_coeffs_indices,
        op_prod_coeffs_indices,
        base_op_indices,
    )
    return state_out_bufs, (op, state_in_bufs)


def _operator_action_bwd(device: jax.Device,
                         batch_size: int,
                         num_state_components: int,
                         purity,
                         other_in_types: tuple[int, ...],
                         other_in_ptrs: tuple[int, ...],
                         other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                         other_out_types: tuple[int, ...],
                         other_out_ptrs: tuple[int, ...],
                         op_term_coeffs_indices: tuple[int, ...],
                         op_prod_coeffs_indices: tuple[int, ...],
                         base_op_indices: tuple[int, ...],
                         op_term_coeff_grad_indices: tuple[int, ...],
                         op_prod_coeff_grad_indices: tuple[tuple[int, int], ...],
                         base_op_grad_indices: tuple[tuple[int, int, int], ...],
                         res: tuple[Operator, tuple[jax.Array, ...]],
                         state_out_adj_bufs: jax.Array | Sequence[jax.Array]
                         ) -> tuple[Operator, tuple[jax.Array, ...]]:
    """
    Backward rule for operator_action.

    Args:
        device: Device for the operator action.
        batch_size: Batch size of the operator action.
        num_state_components: Number of state components.
        other_in_types: Non-differentiable argument from forward.
        other_in_ptrs: Non-differentiable argument from forward.
        op_term_coeffs_indices: Non-differentiable argument from forward.
        op_prod_coeffs_indices: Non-differentiable argument from forward.
        base_op_indices: Non-differentiable argument from forward.
        op_term_coeff_grad_indices: Non-differentiable argument from forward.
        op_prod_coeff_grad_indices: Non-differentiable argument from forward.
        base_op_grad_indices: Non-differentiable argument from forward.
        res: Residuals from forward pass.
        state_out_adj_bufs: Data buffers of the output state adjoint.
    """
    logger.info("Calling _operator_action_bwd")

    op, state_in_bufs = res

    # Prepare library context for backward operator action
    if is_vmap_traced(state_in_bufs[0]):
        state_shape = state_in_bufs[0].val.shape
    else:
        state_shape = state_in_bufs[0].shape
    state_ctx = CudensitymatContext.get_state_context(purity, state_shape, batch_size, state_in_bufs[0].dtype)
    state_ctx.create_adjoint_buffers()

    # Process input argument.
    if isinstance(state_out_adj_bufs, jax.Array):
        state_out_adj_bufs = (state_out_adj_bufs,)
    else:
        state_out_adj_bufs = tuple(state_out_adj_bufs)

    if len(state_in_bufs) != len(state_out_adj_bufs):
        raise ValueError("state_in_bufs and state_out_adj_bufs must have the same number of components.")

    op_grad, state_in_adj_bufs = operator_action_backward_diff_prim(
        op,
        state_in_bufs,
        state_out_adj_bufs,
        device,
        batch_size,
        num_state_components,
        state_shape,
        purity,
        other_in_types,
        other_in_ptrs,
        other_out_shape_dtypes,
        other_out_types,
        other_out_ptrs,
        op_term_coeffs_indices,
        op_prod_coeffs_indices,
        base_op_indices,
        op_term_coeff_grad_indices,
        op_prod_coeff_grad_indices,
        base_op_grad_indices,
    )

    if len(state_in_adj_bufs) == 1:
        state_in_adj_bufs = state_in_adj_bufs[0]

    return op_grad, state_in_adj_bufs


_operator_action.defvjp(_operator_action_fwd, _operator_action_bwd)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

# Note: Project metadata (name, version, dependencies, classifiers) is
# declared in pyproject.toml, which is generated from pyproject.toml.template
# by configure.sh.  This file only contains build mechanics (CMake extension
# compilation, wheel packaging, and package layout).

import os
import sys
import shutil
import pathlib
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.bdist_wheel import bdist_wheel


# ---------------------------------------------------------------------------
# Guard: verify that configure.sh has been run before building.
# ---------------------------------------------------------------------------
_pyproject = os.path.join(os.path.dirname(os.path.abspath(__file__)), "pyproject.toml")
if os.path.exists(_pyproject):
    with open(_pyproject) as _f:
        if "[project]" not in _f.read():
            raise RuntimeError(
                "pyproject.toml does not contain [project] metadata.\n"
                "Run configure.sh first:\n"
                "  CUDA_PATH=/usr/local/cuda bash configure.sh"
            )
else:
    raise RuntimeError(
        "pyproject.toml not found.\n"
        "Run configure.sh first:\n"
        "  CUDA_PATH=/usr/local/cuda bash configure.sh"
    )


class CMakeExtension(Extension):
    def __init__(self, name: str, source_dir: str = ""):
        super().__init__(name, sources=[])
        self.source_dir = os.path.abspath(source_dir)

class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = []

        additional_cmake_args = os.environ.get("CMAKE_ARGS", "")
        if additional_cmake_args:
            cmake_args += additional_cmake_args.split()

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
        cmake_args += [f"-DPython3_EXECUTABLE={sys.executable}"]
        build_args += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ["cmake", ext.source_dir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", ".", "--verbose"] + build_args, cwd=self.build_temp
        )
        dso_dir = os.path.join(extdir, "cuquantum/lib")
        os.makedirs(dso_dir, exist_ok=True)
        for dso in pathlib.Path(self.build_temp).rglob("*.so"):
            shutil.copy2(dso, dso_dir)

class CustomBdistWheel(bdist_wheel):
    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False

setup(
    ext_modules=[
        CMakeExtension("cppsrc", source_dir="cuquantum/densitymat/jax/cppsrc"),
    ],
    cmdclass=dict(
        build_ext=CMakeBuild,
        bdist_wheel=CustomBdistWheel
    ),
    packages=[
        "cuquantum.densitymat.jax",
        "cuquantum.densitymat.jax.pysrc"
    ],
    package_data={
        "cuquantum.densitymat.jax": ["**/*.so"]
    },
    exclude_package_data={
        "cuquantum.densitymat.jax": ["cppsrc/*"],
        "": ["tests/*"]
    },
    include_package_data=True,
    zip_safe=False
)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Operator class in cuDensityMat.
"""

import ctypes
import logging
from collections.abc import Sequence

import cupy as cp
import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm

from .operator_term import OperatorTerm
from ..utils import (
    get_scalar_assignment_callback,
    get_empty_scalar_callback,
    get_scalar_gradient_attachment_callback,
    get_random_odd_pointer_and_object,
    detect_ad_traced_object,
    is_vmap_traced,
)


@jax.tree_util.register_pytree_node_class
class Operator:
    """
    PyTree class for cuDensityMat's operator.
    """

    logger = logging.getLogger("cudensitymat-jax.Operator")

    def __init__(self, dims: Sequence[int]) -> None:
        """
        Initialize an Operator object.

        Args:
            dims: Hilbert space dimensions.
        """
        # Attribute set from constructor.
        self.dims: tuple[int, ...] = tuple(dims)

        # Attributes for arguments in append.
        self.op_terms: list[OperatorTerm] = []
        self.duals: list[bool] = []
        self.coeffs: list[jax.Array] = []

        # Attributes inferred from multiple append calls.
        self.batch_sizes: list[int] = []  # keep track of batch sizes of all operator terms
        self.batch_size: int = 1
        self.dtype: jnp.dtype | None = None

        # Internal attributes from interfacing to cuDensityMat.
        self._ptr: int | None = None

        # Attributes for handling JIT tracing.
        self._coeff_ptrs: list[int] = []
        self._coeff_ptr_objs: list[ctypes.c_short | None] = []  # Keep ctypes objects alive

        self._coeff_grad_ptrs: list[int] = []
        self._coeff_grad_ptr_objs: list[ctypes.c_short | None] = []  # Keep ctypes objects alive

        # Attributes for handling gradients.
        self._coeff_requires_grads: list[bool] = []
        self._coeff_callbacks = []
        self._coeff_grad_callbacks = []
        self._op_term_ids: list[int] = []
        self._total_coeffs_ptrs: list[int] = []
        self._total_coeffs_ptr_objs: list[ctypes.c_short | None] = []  # Keep ctypes objects alive

    def tree_flatten(self):
        """
        Flatten the operator PyTree.
        """
        children = (self.op_terms, self.coeffs)
        aux_data = (
            self.dims,
            self.duals,
            self.batch_size,
            self.batch_sizes,
            self.dtype,
            self._ptr,
            self._coeff_ptrs,
            self._coeff_ptr_objs,
            self._coeff_grad_ptrs,
            self._coeff_grad_ptr_objs,
            self._coeff_requires_grads,
            self._coeff_callbacks,
            self._coeff_grad_callbacks,
            self._op_term_ids,
            self._total_coeffs_ptrs,
            self._total_coeffs_ptr_objs,
        )
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the operator PyTree.
        """
        inst = cls.__new__(cls)
        inst.op_terms, inst.coeffs = children
        (
            inst.dims,
            inst.duals,
            inst.batch_size,
            inst.batch_sizes,
            inst.dtype,
            inst._ptr,
            inst._coeff_ptrs,
            inst._coeff_ptr_objs,
            inst._coeff_grad_ptrs,
            inst._coeff_grad_ptr_objs,
            inst._coeff_requires_grads,
            inst._coeff_callbacks,
            inst._coeff_grad_callbacks,
            inst._op_term_ids,
            inst._total_coeffs_ptrs,
            inst._total_coeffs_ptr_objs,
        ) = aux_data
        return inst

    @property
    def in_axes(self) -> "Operator":
        """
        Return the in_axes PyTree spec for vmapping over the batch dimension.
        """
        in_axes_op_terms = [op_term.in_axes for op_term in self.op_terms]
        in_axes_coeffs = []
        for coeff in self.coeffs:
            if is_vmap_traced(coeff) or len(coeff) > 1:
                in_axes_coeffs.append(0)
            else:
                in_axes_coeffs.append(None)

        _, aux_data = self.tree_flatten()
        return type(self).tree_unflatten(aux_data, (in_axes_op_terms, in_axes_coeffs))
    
    def _copy(self) -> "Operator":
        """
        Internal method to copy the operator for VJP backward pass.
        """
        op = type(self).__new__(type(self))

        op.op_terms = [op_term._copy() for op_term in self.op_terms]
        op.coeffs = [jnp.copy(c) for c in self.coeffs]

        op.dims = self.dims
        op.duals = self.duals.copy()
        op.batch_size = self.batch_size
        op.batch_sizes = self.batch_sizes.copy()
        op.dtype = self.dtype
        op._ptr = self._ptr
        op._coeff_ptrs = self._coeff_ptrs.copy()
        op._coeff_ptr_objs = self._coeff_ptr_objs.copy()
        op._coeff_grad_ptrs = self._coeff_grad_ptrs.copy()
        op._coeff_grad_ptr_objs = self._coeff_grad_ptr_objs.copy()
        op._coeff_requires_grads = self._coeff_requires_grads.copy()
        op._coeff_callbacks = self._coeff_callbacks.copy()
        op._coeff_grad_callbacks = self._coeff_grad_callbacks.copy()
        op._op_term_ids = self._op_term_ids.copy()
        op._total_coeffs_ptrs = self._total_coeffs_ptrs.copy()
        op._total_coeffs_ptr_objs = self._total_coeffs_ptr_objs.copy()
        return op

    def _check_and_set_dtype(self, op_term: OperatorTerm) -> None:
        """
        Check if the operator term has the same data type as the operator.
        """
        if op_term.dtype is not None:  # for empty operator term, skip the check.
            if self.dtype is None:
                # If the data type is not set, set it to the data type of the first operator term.
                self.dtype = op_term.dtype
            else:
                # If the data type is set, check if the operator term has the same data type as the operator.
                if op_term.dtype != self.dtype:
                    raise ValueError("All operator terms must have the same data type.")

    def _check_and_set_batch_size(self, op_term: OperatorTerm, coeff: jax.Array) -> None:
        """
        Check if the operator term and coefficient batch sizes are consistent.
        """
        # Possibly update the batch size of this operator and check consistency.
        batch_size = max(op_term.batch_size, len(coeff))
        if self.batch_size == 1:
            self.batch_size = batch_size
        else:
            if batch_size not in (1, self.batch_size):
                raise ValueError("Batch size in this operator term does not match batch size of this operator.")

    def append(self,
               op_term: OperatorTerm,
               dual: bool = False,
               coeff: float | complex | jax.Array = 1.0,
               ) -> None:
        """
        Append an operator term to an operator.

        Args:
            op_term: Operator term to be appended.
            dual: Duality of the operator term.
            coeff: Non-batched coefficient or batched coefficients of the operator term.
            coeff_requires_grad: Whether the coefficients require gradient.
        """
        # TODO: Instead of an explicit coeff_requires_grad argument, this should be detected
        # automatically from the trace stack.

        if self._ptr is not None:
            raise RuntimeError("Cannot modify operator after it has been used in an operator action.")

        # coeff is converted to a length-1 array if it is a Python scalar.
        if (
            isinstance(coeff, (float, complex)) or
            (isinstance(coeff, jax.Array) and coeff.ndim == 0)  # scalar but traced
        ):
            coeff = jnp.array([coeff], dtype=jnp.complex128)
        elif isinstance(coeff, jax.Array) and coeff.ndim > 0:
            if coeff.dtype != jnp.complex128:
                raise ValueError("Coefficient must be of type complex128.")
        else:
            raise ValueError("Coefficient must be a float, complex, or jax.Array.")

        # Attributes from function arguments.
        self.op_terms.append(op_term)
        self.duals.append(dual)
        self.coeffs.append(coeff)

        # Set batch size and dtype.
        self.batch_sizes.append(len(coeff))
        self._check_and_set_batch_size(op_term, coeff)  # setting batch size of the operator
        self._check_and_set_dtype(op_term)

        # Internal attributes.
        self._coeff_callbacks.append(None)
        self._coeff_ptrs.append(0)
        self._coeff_ptr_objs.append(None)

        self._coeff_requires_grads.append(None)
        self._coeff_grad_callbacks.append(None)
        self._coeff_grad_ptrs.append(0)
        self._coeff_grad_ptr_objs.append(None)
        self._total_coeffs_ptrs.append(0)
        self._total_coeffs_ptr_objs.append(None)
        self._op_term_ids.append(id(op_term))

    def __getitem__(self, index: int) -> OperatorTerm:
        """
        Get an operator term from the operator.
        """
        return self.op_terms[index]

    def _create(self, handle):
        """
        Create opaque handle to the operator.
        """
        # Create a dictionary to map from the original op_term_id to the first index of
        # the op_term in the operator. The original op_term_id need to be used since id(op_term)
        # changes when JAX flattens and unflattens the PyTrees.
        id_to_first_index = {}
        for i, op_term_id in enumerate(self._op_term_ids):
            if op_term_id not in id_to_first_index:
                id_to_first_index[op_term_id] = i

        # Only create the opaque handles to the unique operator terms.
        for i in id_to_first_index.values():
            self.op_terms[i]._create(handle)

        # Create the current operator.
        if self._ptr is None:
            self._ptr = cudm.create_operator(handle, len(self.dims), self.dims)
            self.logger.debug(f"Created operator at {hex(self._ptr)}")
            # Keep batched coefficient buffers alive so C API pointers remain valid.
            self._batch_coeff_arrs = []

            for i in range(len(self.op_terms)):
                # Detect if the coefficient requires gradient and assign callback, gradient callback,
                # temporary coefficient pointer and object.
                self._coeff_requires_grads[i] = detect_ad_traced_object(self.coeffs[i])
                if not is_vmap_traced(self.coeffs[i]) and len(self.coeffs[i]) == 1:
                    # Traced scalars need to be passed through an intermediate memory slot.
                    self._coeff_callbacks[i] = get_scalar_assignment_callback(self.coeffs[i])
                    self._coeff_ptrs[i] = self._coeff_callbacks[i].callback.coeff.data.ptr

                    # If gradient is computed on the coefficient, assign gradient callback and pointer.
                    if self._coeff_requires_grads[i]:
                        self._coeff_grad_callbacks[i] = get_scalar_gradient_attachment_callback(self.coeffs[i])
                        self._coeff_grad_ptrs[i] = self._coeff_grad_callbacks[i].callback.scalar_grad.data.ptr

                else:
                    if is_vmap_traced(self.coeffs[i]):
                        coeff_shape = self.coeffs[i].val.shape
                        coeff_dtype = self.coeffs[i].val.dtype
                    else:
                        coeff_shape = self.coeffs[i].shape
                        coeff_dtype = self.coeffs[i].dtype
                    static_coeff_buf = cp.ones(coeff_shape, dtype=coeff_dtype)
                    self._coeff_ptrs[i] = static_coeff_buf.data.ptr
                    self._coeff_ptr_objs[i] = static_coeff_buf

                    # For batched coefficients, callback is required by the API but is a no-op.
                    self._coeff_callbacks[i] = get_empty_scalar_callback()
                    self._total_coeffs_ptrs[i], self._total_coeffs_ptr_objs[i] = get_random_odd_pointer_and_object()

                    if self._coeff_requires_grads[i]:
                        self._coeff_grad_callbacks[i] = get_scalar_gradient_attachment_callback(self.coeffs[i])
                        self._coeff_grad_ptrs[i] = self._coeff_grad_callbacks[i].callback.scalar_grad.data.ptr

                if self.batch_sizes[i] == 1:
                    cudm.operator_append_term(
                        handle,
                        self._ptr,
                        self.op_terms[id_to_first_index[self._op_term_ids[i]]]._ptr,
                        self.duals[i],
                        1.0,  # coefficient, to be updated to the real coefficient by callback
                        self._coeff_callbacks[i],
                        self._coeff_grad_callbacks[i],
                    )
                else:
                    # This is only needed when is_vmap_traced(self.coeffs[i]) or len(self.coeffs[i]) > 1,
                    # and when _self._coeff_requires_grads[i] is True.
                    cudm.operator_append_term_batch(
                        handle,
                        self._ptr,
                        self.op_terms[id_to_first_index[self._op_term_ids[i]]]._ptr,
                        self.duals[i],
                        self.batch_sizes[i],
                        self._coeff_ptrs[i],
                        self._total_coeffs_ptrs[i],
                        self._coeff_callbacks[i],
                        self._coeff_grad_callbacks[i],
                    )
            self.logger.debug(f"Appended operator terms to operator at {hex(self._ptr)}")
    
    def _destroy(self):
        """
        Destroy opaque handle to the operator.
        """
        if self._ptr is not None:
            # Destroy the current operator.
            cudm.destroy_operator(self._ptr)
            self.logger.debug(f"Destroyed operator at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Operator action primitive.
"""

import logging

import jax
import jax.numpy as jnp
from jax.interpreters import mlir

from cuquantum.bindings import cudensitymat as cudm

from ..utils import fuse_batched_inputs, unfuse_batched_outputs
from .base import BasePrimitive, register_primitive
from .context import CudensitymatContext
from .operator import Operator


logger = logging.getLogger("cudensitymat-jax.operator_action")


class OperatorActionPrimitive(BasePrimitive):
    """
    JAX primitive for operator action.
    """

    name = "operator_action"
    inner_multiple_results = True
    outer_multiple_results = True
    inner_primitive = None
    outer_primitive = None

    logger = logging.getLogger("cudensitymat-jax.OperatorActionPrimitive")

    @staticmethod
    def abstract(*in_buf_avals: list[jax.core.ShapedArray],
                 device: jax.Device,
                 batch_size: int,
                 num_state_components: int,
                 other_in_types: tuple[int, ...],
                 other_in_ptrs: tuple[int, ...],
                 other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                 other_out_types: tuple[int, ...],
                 other_out_ptrs: tuple[int, ...],
                 op_ptr: int,
                 state_shape: tuple[int, ...],
                 purity: cudm.StatePurity,
                 ) -> tuple[jax.core.ShapedArray, ...]:
        """
        Abstract evaluation of the inner primitive of operator action.
        """
        OperatorActionPrimitive.logger.info("Calling abstract evaluation of the inner primitive")

        dtype = in_buf_avals[0].dtype
        op_ctx = CudensitymatContext.get_operator_context(op_ptr)
        state_ctx = CudensitymatContext.get_state_context(purity, state_shape, batch_size, dtype)

        # Create abstract arrays for the output state buffers.
        state_out_buf_avals = [
            jax.core.ShapedArray(in_buf_avals[i].shape, in_buf_avals[i].dtype)
            for i in range(num_state_components)
        ]

        other_out_buf_avals = [
            jax.core.ShapedArray(other_out_shape_dtypes[i].shape, other_out_shape_dtypes[i].dtype)
            for i in range(len(other_out_shape_dtypes))
        ]

        # Obtain workspace limit and stream from the device.
        workspace_limit = device.memory_stats()['bytes_limit'] - device.memory_stats()['bytes_in_use']
        stream = device.get_stream_for_external_ready_events()

        # Prepare operator action.
        cudm.operator_prepare_action(
            CudensitymatContext._handle,
            op_ctx._operator,
            state_ctx._state_in,
            state_ctx._state_out,
            op_ctx._compute_type,
            workspace_limit,
            CudensitymatContext._workspace_desc,
            stream)

        # Query the required buffer size for the workspace.
        required_buffer_size = cudm.workspace_get_memory_size(
            CudensitymatContext._handle,
            CudensitymatContext._workspace_desc,
            cudm.Memspace.DEVICE,
            cudm.WorkspaceKind.WORKSPACE_SCRATCH)

        # Create abstract workspace array.
        # NOTE: Memory buffers from cudaMalloc is automatically 256-aligned, which is not 
        # the case for JAX. 255 is added to the buffer size to ensure workspace is 256-aligned.
        workspace_aval = jax.core.ShapedArray((required_buffer_size + 255,), jnp.uint8)
        return workspace_aval, *state_out_buf_avals, *other_out_buf_avals

    @staticmethod
    def outer_abstract(*args, **kwargs):
        """
        Abstract evaluation of the outer primitive of operator action.
        """
        OperatorActionPrimitive.logger.info("Calling abstract evaluation of the outer primitive")
        _, *out = OperatorActionPrimitive.abstract(*args, **kwargs)
        return out

    @staticmethod
    def lowering(ctx: mlir.LoweringRuleContext,
                 *in_bufs: tuple[mlir.ir.OpResult, ...],
                 device: jax.Device,
                 batch_size: int,
                 num_state_components: int,
                 other_in_types: tuple[int, ...],
                 other_in_ptrs: tuple[int, ...],
                 other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                 other_out_types: tuple[int, ...],
                 other_out_ptrs: tuple[int, ...],
                 op_ptr: int,
                 state_shape: tuple[int, ...],
                 purity: cudm.StatePurity,
                 ) -> tuple[mlir.ir.OpResult, ...]:
        """
        Lowering rule of the operator action primitive.
        """
        OperatorActionPrimitive.logger.info("Calling lowering rule")

        dtype = ctx.avals_in[0].dtype
        state_ctx = CudensitymatContext.get_state_context(purity, state_shape, batch_size, dtype)

        # Revert indices in input and output states. Note the layout is specified as
        # minor-to-major axis order. For 0-d tensors use () (StableHLO requires empty layout).
        def _layout_for_ndim(ndim):
            return () if ndim == 0 else tuple(range(1, ndim)) + (0,)

        operand_layouts = [None] * len(ctx.avals_in)
        for i in range(len(ctx.avals_in)):
            # 0 (the batch dimension) is the most major axis in input buffers and also when passed
            # to the cuQuantum library. Other dimensions (Hilbert space modes) need to be reversed.
            operand_layouts[i] = _layout_for_ndim(ctx.avals_in[i].ndim)

        result_layouts = [None] * len(ctx.avals_out)
        for i in range(len(ctx.avals_out)):
            result_layouts[i] = _layout_for_ndim(ctx.avals_out[i].ndim)

        # Lower to the XLA FFI handler.
        outputs = jax.ffi.ffi_lowering(
            OperatorActionPrimitive.name,
            operand_layouts=operand_layouts,
            result_layouts=result_layouts
        )(
            ctx,
            *in_bufs,
            other_in_types=mlir.dense_int_elements(other_in_types),
            other_in_ptrs=mlir.dense_int_elements(other_in_ptrs),
            batch_size=batch_size,
            num_state_components=num_state_components,
            other_out_types=mlir.dense_int_elements(other_out_types),
            other_out_ptrs=mlir.dense_int_elements(other_out_ptrs),
            handle=CudensitymatContext._handle,
            operator=op_ptr,
            state_in=state_ctx._state_in,
            state_out=state_ctx._state_out
        )
        return outputs

    @staticmethod
    def impl(*args, **kwargs):
        """
        Primal evaluation of the operator action primitive.
        """
        OperatorActionPrimitive.logger.info("Calling primal evaluation")

        assert OperatorActionPrimitive.inner_primitive is not None
        _, *out = OperatorActionPrimitive.inner_primitive.bind(*args, **kwargs)
        return out

    @staticmethod
    def batcher(batched_args, batch_dims, **kwargs):
        """
        Batching rule of the operator action primitive.
        """
        OperatorActionPrimitive.logger.info("Calling batcher")

        num_state_components = kwargs['num_state_components']

        # Fuse pivot batch axis with vmap axis.
        fused_inputs, batch_sizes, vmap_sizes = fuse_batched_inputs(
            (0,) * len(batched_args),
            batched_args,
            batch_dims,
            num_state_components,
        )

        # State context key uses fused state shape and batch size.
        kwargs = dict(kwargs)
        kwargs['state_shape'] = tuple(fused_inputs[0].shape)
        kwargs['batch_size'] = int(fused_inputs[0].shape[0])

        # Invoke outer primitive.
        outputs = OperatorActionPrimitive.outer_primitive.bind(*fused_inputs, **kwargs)

        # Unfuse pivot batch axis and vmap axis.
        outputs[:num_state_components] = unfuse_batched_outputs(
            outputs[:num_state_components],
            (0,) * num_state_components,
            batch_sizes,
            vmap_sizes,
        )
        return outputs, (0,) * len(outputs)

register_primitive(OperatorActionPrimitive)


def operator_action_prim(op: Operator,
                         state_in_bufs: tuple[jax.Array, ...],
                         device: jax.Device,
                         batch_size: int,
                         num_state_components: int,
                         purity: cudm.StatePurity,
                         other_in_types: tuple[int, ...],
                         other_in_ptrs: tuple[int, ...],
                         other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                         other_out_types: tuple[int, ...],
                         other_out_ptrs: tuple[int, ...],
                         op_term_coeffs_indices: tuple[int, ...],
                         op_prod_coeffs_indices: tuple[tuple[int, int], ...],
                         base_op_indices: tuple[tuple[int, int, int], ...],
                         ) -> tuple[jax.Array, ...]:
    """
    Function wrapper around OperatorActionPrimitive.
    """
    logger.info("Calling operator_action_prim")

    other_in_bufs = []

    # Extract buffers using the same index structure as operator_action.
    for i in op_term_coeffs_indices:
        other_in_bufs.append(op.coeffs[i])

    for i, j in op_prod_coeffs_indices:
        other_in_bufs.append(op[i].coeffs[j])

    for i, j, k in base_op_indices:
        other_in_bufs.append(op[i][j][k].data)

    state_shape = tuple(state_in_bufs[0].shape)
    out = OperatorActionPrimitive.outer_primitive.bind(
        *state_in_bufs,
        *other_in_bufs,
        device=device,
        batch_size=batch_size,
        num_state_components=num_state_components,
        other_in_types=tuple(other_in_types),
        other_in_ptrs=tuple(other_in_ptrs),
        other_out_shape_dtypes=other_out_shape_dtypes,
        other_out_types=other_out_types,
        other_out_ptrs=other_out_ptrs,
        op_ptr=op._ptr,
        state_shape=state_shape,
        purity=purity,
    )

    state_out_bufs = out[:num_state_components]
    return state_out_bufs


class OperatorActionBackwardDiffPrimitive(BasePrimitive):
    """
    JAX primitive for operator action backward differentiation.
    """

    name = "operator_action_backward_diff"
    inner_multiple_results = True
    outer_multiple_results = True
    inner_primitive = None
    outer_primitive = None

    logger = logging.getLogger("cudensitymat-jax.OperatorActionBackwardDiffPrimitive")

    @staticmethod
    def abstract(*in_buf_avals: list[jax.core.ShapedArray],
                 device: jax.Device,
                 batch_size: int,
                 num_state_components: int,
                 other_in_types: tuple[int, ...],
                 other_in_ptrs: tuple[int, ...],
                 other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                 other_out_types: tuple[int, ...],
                 other_out_ptrs: tuple[int, ...],
                 op_ptr: int,
                 state_shape: tuple[int, ...],
                 purity: cudm.StatePurity,
                 ) -> tuple[jax.core.ShapedArray, ...]:
        """
        Abstract evaluation of the inner primitive of operator action backward differentiation.
        """
        OperatorActionBackwardDiffPrimitive.logger.info("Calling abstract evaluation of the inner primitive")

        dtype = in_buf_avals[0].dtype
        op_ctx = CudensitymatContext.get_operator_context(op_ptr)
        state_ctx = CudensitymatContext.get_state_context(purity, state_shape, batch_size, dtype)

        # Extract state input adjoint buffer shapes from state input buffers.
        state_in_adj_buf_avals = [
            jax.core.ShapedArray(in_buf_avals[i].shape, in_buf_avals[i].dtype)
            for i in range(num_state_components)
        ]

        other_out_buf_avals = [
            jax.core.ShapedArray(other_out_shape_dtypes[i].shape, other_out_shape_dtypes[i].dtype)
            for i in range(len(other_out_shape_dtypes))
        ]

        # Obtain workspace limit and stream from the device.
        workspace_limit = device.memory_stats()['bytes_limit'] - device.memory_stats()['bytes_in_use']
        stream = device.get_stream_for_external_ready_events()

        # Prepare operator action backward differentiation.
        cudm.operator_prepare_action_backward_diff(
            CudensitymatContext._handle,
            op_ctx._operator,
            state_ctx._state_in,
            state_ctx._state_out_adj,
            op_ctx._compute_type,
            workspace_limit,
            CudensitymatContext._workspace_desc,
            stream)

        # Query the required buffer size for the workspace.
        required_buffer_size = cudm.workspace_get_memory_size(
            CudensitymatContext._handle,
            CudensitymatContext._workspace_desc,
            cudm.Memspace.DEVICE,
            cudm.WorkspaceKind.WORKSPACE_SCRATCH)

        # Create abstract workspace array.
        # NOTE: Memory buffers from cudaMalloc is automatically 256-aligned, which is not 
        # the case for JAX. 255 is added to the buffer size to ensure workspace is 256-aligned.
        workspace_aval = jax.core.ShapedArray((required_buffer_size + 255,), jnp.uint8)
        return workspace_aval, *state_in_adj_buf_avals, *other_out_buf_avals

    @staticmethod
    def outer_abstract(*args, **kwargs):
        """
        Abstract evaluation of the outer primitive of operator action backward differentiation.
        """
        OperatorActionBackwardDiffPrimitive.logger.info("Calling abstract evaluation of the outer primitive")
        _, *out = OperatorActionBackwardDiffPrimitive.abstract(*args, **kwargs)
        return out

    @staticmethod
    def lowering(ctx: mlir.LoweringRuleContext,
                 *in_bufs: tuple[mlir.ir.OpResult, ...],
                 device: jax.Device,
                 batch_size: int,
                 num_state_components: int,
                 other_in_types: tuple[int, ...],
                 other_in_ptrs: tuple[int, ...],
                 other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                 other_out_types: tuple[int, ...],
                 other_out_ptrs: tuple[int, ...],
                 op_ptr: int,
                 state_shape: tuple[int, ...],
                 purity: cudm.StatePurity,
                 ) -> tuple[mlir.ir.OpResult, ...]:
        """
        Lowering rule of the operator action backward differentiation primitive.
        """
        OperatorActionBackwardDiffPrimitive.logger.info("Calling lowering rule")

        dtype = ctx.avals_in[0].dtype
        state_ctx = CudensitymatContext.get_state_context(purity, state_shape, batch_size, dtype)

        # Revert indices in input and output states. Note the layout is specified as
        # minor-to-major axis order. For 0-d tensors use () (StableHLO requires empty layout).
        def _layout_for_ndim(ndim):
            return () if ndim == 0 else tuple(range(1, ndim)) + (0,)

        operand_layouts = [None] * len(ctx.avals_in)
        for i in range(len(ctx.avals_in)):
            # 0 (the batch dimension) is the most major axis in input buffers and also when passed
            # to the cuQuantum library. Other dimensions (Hilbert space modes) need to be reversed.
            operand_layouts[i] = _layout_for_ndim(ctx.avals_in[i].ndim)

        result_layouts = [None] * len(ctx.avals_out)
        # for i in range(len(ctx.avals_out)):
        for i in range(num_state_components):  # XXX
            result_layouts[i] = _layout_for_ndim(ctx.avals_out[i].ndim)

        # Lower to the XLA FFI handler.
        outputs = jax.ffi.ffi_lowering(
            OperatorActionBackwardDiffPrimitive.name,
            operand_layouts=operand_layouts,
            result_layouts=result_layouts
        )(
            ctx,
            *in_bufs,
            other_in_types=mlir.dense_int_elements(other_in_types),
            other_in_ptrs=mlir.dense_int_elements(other_in_ptrs),
            batch_size=batch_size,
            num_state_components=num_state_components,
            other_out_types=mlir.dense_int_elements(other_out_types),
            other_out_ptrs=mlir.dense_int_elements(other_out_ptrs),
            handle=CudensitymatContext._handle,
            operator=op_ptr,
            state_in=state_ctx._state_in,
            state_out_adj=state_ctx._state_out_adj,
            state_in_adj=state_ctx._state_in_adj
        )
        return outputs

    @staticmethod
    def impl(*args, **kwargs):
        """
        Primal evaluation of the operator action backward differentiation primitive.
        """
        OperatorActionBackwardDiffPrimitive.logger.info("Calling primal evaluation")

        assert OperatorActionBackwardDiffPrimitive.inner_primitive is not None
        _, *out = OperatorActionBackwardDiffPrimitive.inner_primitive.bind(*args, **kwargs)
        return out

    @staticmethod
    def batcher(batched_args, batch_dims, **kwargs):
        """
        Batching rule of the operator action backward differentiation primitive.
        """
        OperatorActionBackwardDiffPrimitive.logger.info("Calling batcher")

        num_state_components = kwargs['num_state_components']

        # Fuse pivot batch axis with vmap axis.
        fused_inputs, batch_sizes, vmap_sizes = fuse_batched_inputs(
            (0,) * len(batched_args),
            batched_args,
            batch_dims,
            num_state_components,
        )

        kwargs = dict(kwargs)
        kwargs['state_shape'] = tuple(fused_inputs[0].shape)
        kwargs['batch_size'] = int(fused_inputs[0].shape[0])

        # Invoke outer primitive.
        outputs = OperatorActionBackwardDiffPrimitive.outer_primitive.bind(
            *fused_inputs, **kwargs)

        # Unfuse pivot batch axis and vmap axis for state adjoint outputs only.
        outputs[:num_state_components] = unfuse_batched_outputs(
            outputs[:num_state_components],
            (0,) * num_state_components,
            batch_sizes,
            vmap_sizes,
        )
        return outputs, (0,) * len(outputs)

register_primitive(OperatorActionBackwardDiffPrimitive)


def operator_action_backward_diff_prim(op: Operator,
                                       state_in_bufs: tuple[jax.Array, ...],
                                       state_out_adj_bufs: tuple[jax.Array, ...],
                                       device: jax.Device,
                                       batch_size: int,
                                       num_state_components: int,
                                       state_shape: tuple[int, ...],
                                       purity: cudm.StatePurity,
                                       other_in_types: tuple[int, ...],
                                       other_in_ptrs: tuple[int, ...],
                                       other_out_shape_dtypes: tuple[jax.ShapeDtypeStruct, ...],
                                       other_out_types: tuple[int, ...],
                                       other_out_ptrs: tuple[int, ...],
                                       op_term_coeffs_indices: tuple[int, ...],
                                       op_prod_coeffs_indices: tuple[tuple[int, int], ...],
                                       base_op_indices: tuple[tuple[int, int, int], ...],
                                       op_term_coeff_grad_indices: tuple[int, ...],
                                       op_prod_coeff_grad_indices: tuple[tuple[int, int], ...],
                                       base_op_grad_indices: tuple[tuple[int, int, int], ...],
                                       ) -> tuple[jax.Array, ...]:
    """
    Wrapper around the outer primitive of OperatorActionBackwardDiffPrimitive.
    """
    logger.info("Calling operator_action_backward_diff_prim")

    other_in_bufs = []

    # Extract buffers using the same index structure as operator_action.
    for i in op_term_coeffs_indices:
        other_in_bufs.append(op.coeffs[i])

    for i, j in op_prod_coeffs_indices:
        other_in_bufs.append(op[i].coeffs[j])

    for i, j, k in base_op_indices:
        other_in_bufs.append(op[i][j][k].data)

    out = OperatorActionBackwardDiffPrimitive.outer_primitive.bind(
        *state_in_bufs,
        *state_out_adj_bufs,
        *other_in_bufs,
        device=device,
        batch_size=batch_size,
        num_state_components=num_state_components,
        state_shape=tuple(state_shape),
        purity=purity,
        other_in_types=other_in_types,
        other_in_ptrs=other_in_ptrs,
        other_out_shape_dtypes=other_out_shape_dtypes,
        other_out_types=other_out_types,
        other_out_ptrs=other_out_ptrs,
        op_ptr=op._ptr,
    )

    state_in_adj_bufs = out[:num_state_components]
    grad_bufs = out[num_state_components:]

    # In _operator_action_bwd, after getting grad_bufs:
    op_grad = op._copy()
    grad_idx = 0

    # Zero out all base operators that require gradient. The backend accumulates the
    # total gradient into a single buffer per unique base operator (_grad_ptr). We only
    # create one output buffer per unique _grad_ptr (deduplicated in operator_action.py),
    # so the first occurrence gets the correct total gradient and all other (duplicate)
    # occurrences must be zero so JAX doesn't overcount when it sums pytree leaves.
    for i, op_term in enumerate(op_grad.op_terms):
        op_grad.coeffs[i] = jnp.zeros_like(op_grad.coeffs[i])
        for j, op_prod in enumerate(op_term.op_prods):
            op_grad[i].coeffs[j] = jnp.zeros_like(op_grad[i].coeffs[j])
            for k, base_op in enumerate(op_prod):
                if base_op.requires_grad:
                    op_grad[i][j][k].data = jnp.zeros_like(op_grad[i][j][k].data)

    for i in op_term_coeff_grad_indices:
        op_grad.coeffs[i] = grad_bufs[grad_idx]
        grad_idx += 1

    for i, j in op_prod_coeff_grad_indices:
        op_grad[i].coeffs[j] = grad_bufs[grad_idx]
        grad_idx += 1

    for i, j, k in base_op_grad_indices:
        op_grad[i][j][k].data = grad_bufs[grad_idx]
        grad_idx += 1

    return op_grad, state_in_adj_bufs
